Package: miniCRAN
Version: 0.3.2
Authors@R: c(
    person("Andrie", "de Vries", role=c("aut", "cre", "cph"), email="apdevries@gmail.com"),
    person("Alex", "Chubaty", role="ctb", email="alex.chubaty@gmail.com"),
    person("Microsoft Corporation", role="cph")
    )
License: GPL-2
Copyright: Andrie de Vries, Microsoft Corporation
Title: Create a Mini Version of CRAN Containing Only Selected Packages
Description: Makes it possible to create an internally consistent
    repository consisting of selected packages from CRAN-like repositories.
    The user specifies a set of desired packages, and 'miniCRAN' recursively
    reads the dependency tree for these packages, then downloads only this
    subset. The user can then install packages from this repository directly,
    rather than from CRAN.  This is useful in production settings, e.g. server
    behind a firewall, or remote locations with slow (or zero) Internet access.
URL: https://github.com/andrie/miniCRAN
BugReports: https://github.com/andrie/miniCRAN/issues
Imports: graphics, httr, methods, stats, tools, utils, igraph,
        assertthat (>= 0.2.0)
Suggests: devtools, knitr, rmarkdown, testthat (>= 2.1.0), covr, withr,
        mockery, testthis, roxygen2, mockr, spelling
LazyData: true
LazyLoad: true
VignetteBuilder: knitr
RoxygenNote: 7.3.1
Encoding: UTF-8
Language: en-US
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-16 09:24:45 UTC; apdev
Author: Andrie de Vries [aut, cre, cph],
  Alex Chubaty [ctb],
  Microsoft Corporation [cph]
Maintainer: Andrie de Vries <apdevries@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-16 10:00:03 UTC
Built: R 4.6.0; ; 2026-04-26 08:47:52 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,pkgDepGraph)
S3method(print,pkgDep)
export(addLocalPackage)
export(addOldPackage)
export(addPackage)
export(addPackageListingGithub)
export(basePkgs)
export(checkVersions)
export(getCranDescription)
export(is.online)
export(makeDepGraph)
export(makeLibrary)
export(makeRepo)
export(oldPackages)
export(pkgAvail)
export(pkgDep)
export(updatePackages)
export(updateRepoIndex)
importFrom(assertthat,"on_failure<-")
importFrom(assertthat,assert_that)
importFrom(graphics,legend)
importFrom(graphics,par)
importFrom(graphics,title)
importFrom(httr,GET)
importFrom(httr,content)
importFrom(httr,stop_for_status)
importFrom(igraph,V)
importFrom(igraph,plot.igraph)
importFrom(methods,is)
importFrom(stats,reshape)
importFrom(stats,setNames)
importFrom(tools,write_PACKAGES)
importFrom(utils,chooseCRANmirror)
importFrom(utils,compareVersion)
importFrom(utils,download.file)
importFrom(utils,download.packages)
importFrom(utils,flush.console)
importFrom(utils,installed.packages)
importFrom(utils,select.list)

tags <- "chron"

# Plot using defaults

if (interactive()){
  pdb <- pkgAvail(
    repos = c(CRAN = getOption("minicran.mran")),
    type = "source"
  )
} else {
  pdb <- cranJuly2014
}


if (interactive()) {
  dg <- makeDepGraph(tags, availPkgs = pdb  , includeBasePkgs = FALSE,
                     suggests = TRUE, enhances = TRUE)
  
  set.seed(43);
  plot(dg)
  
  
  # Move edge legend to top left
  set.seed(42);
  plot(dg, legendPosition = c(-1, 1))
  
  # Change font size and shape size
  set.seed(42);
  plot(dg, legendPosition = c(-1, 1), vertex.size = 20,  cex = 0.5)
  
  
  # Move vertex legend to top right
  set.seed(42);
  plot(dg, legendPosition = c(1, 1), vertex.size = 20,  cex = 0.5)
}

## ----make-repo-1--------------------------------------------------------------
library("miniCRAN")

# define CRAN mirror
mirror <- c(CRAN = "https://cloud.r-project.org")

# Specify list of packages to download
pkgs <- c("foreach")
pkgList <- pkgDep(pkgs, repos = mirror, type = "source", suggests = FALSE, 
                  availPkgs = cranJuly2014)
pkgList

## ----make-repo-2, eval=FALSE--------------------------------------------------
# # Create temporary folder for miniCRAN
# dir.create(pth <- file.path(tempdir(), "miniCRAN"))
# 
# # Make repo for source and win.binary
# makeRepo(pkgList, path = pth, repos = mirror, type = c("source", "win.binary"))

## ----make-repo-3, eval=FALSE--------------------------------------------------
# # List all files in miniCRAN
# list.files(pth, recursive = TRUE, full.names = FALSE)

## ----make-repo-4, eval=FALSE--------------------------------------------------
# # Check for available packages
# pkgAvail(repos = pth, type = "win.binary")[, c(1:3, 5)]

## ----make-repo-5, eval=FALSE--------------------------------------------------
# install.packages(pkgs,
#                  repos = paste0("file:///", pth),
#                  type = "source")

## ----addto-repo-new-1, eval=FALSE---------------------------------------------
# # Add new packages (from CRAN) to the miniCRAN repo
# addPackage("Matrix", path = pth, repos = mirror, type = c("source", "win.binary"))
# pkgAvail(repos = pth, type = "win.binary")[, c(1:3, 5)]

## ----addto-repo-old-1, eval=FALSE---------------------------------------------
# # create a data frame with the package and version info
# oldVers <- data.frame(
#   package = c("foreach", "codetools", "iterators"),
#   version = c("1.4.0", "0.2-7", "1.0.5"),
#   stringsAsFactors = FALSE
# )
# 
# # download old source package version and create repo index
# addOldPackage(pkgList, path = pth, vers = oldVers$version, repos = mirror, type = "source")

## ----addto-repo-old-2, eval=FALSE---------------------------------------------
# # List package versions in the miniCRAN repo (produces warning about duplicates)
# pkgVersionsSrc <- checkVersions(pkgList, path = pth, type = "source")
# pkgVersionsBin <- checkVersions(pkgList, path = pth, type = "win.binary")
# 
# # After inspecting package versions, remove old versions
# basename(pkgVersionsSrc) # duplicate versions
# basename(pkgVersionsBin)
# 
# file.remove(pkgVersionsSrc[c(2,4,6)])
# 
# # rebuild the package index after removing duplicate package versions
# updateRepoIndex(pth, type = c("source", "win.binary"))

## ----addto-repo-old-3, eval=FALSE---------------------------------------------
# pkgAvail(pth, type = "source")[, c(1:3, 5)] # contains the old versions
# pkgAvail(pth, type = "win.binary")[, c(1:3, 5)] # contains the current versions

## ----update-repo-1, eval=FALSE------------------------------------------------
# # Check if updated packages are available
# oldPackages(path = pth, repos = mirror, type = "source")[, 1:3] # should need update
# oldPackages(path = pth, repos = mirror, type = "win.binary")[, 1:3] # should be current

## ----update-repo-2, eval=FALSE------------------------------------------------
# # Update available packages
# updatePackages(path = pth, repos = mirror, type = "source", ask = FALSE) # should need update
# updatePackages(path = pth, repos = mirror, type = "win.binary", ask = FALSE) # should be current

## ----cleanup, include=FALSE, eval=FALSE---------------------------------------
# # Delete temporary folder
# unlink(pth, recursive = TRUE)

if (interactive()) {
  pkgDep(pkg = c("ggplot2", "plyr", "reshape2"),
         repos = c(CRAN = "https://cloud.r-project.org")
  )
  
  pdb <- cranJuly2014
  pdb <- pkgAvail(repos = c(CRAN = getOption("minicran.mran")))
  
  pkgDep(pkg = c("ggplot2", "plyr", "reshape2"), pdb)
  
}

### `oldPackages` and `updatePackages` require an existing miniCRAN repo

# Specify list of packages to download
mirror <- c(CRAN = "https://cloud.r-project.org")
pkgs <- c("foreach")

pdb <- cranJuly2014

if (interactive()) {
  pdb <- pkgAvail(repos = mirror, type = "source")
  
  pkgList <- pkgDep(pkgs, availPkgs = pdb, repos = mirror, type = "source", suggests = FALSE)
  pkgList
  
  # Create temporary folder for miniCRAN
  dir.create(pth <- file.path(tempdir(), "miniCRAN"))
  
  # create the miniCRAN directory structure but only add bin files
  makeRepo(pkgList, path = pth, repos = mirror, type = "source", download = FALSE)
  makeRepo(pkgList, path = pth, repos = mirror, type = "win.binary", download = TRUE)
  
  # download old source package version and create repo index
  oldVers <- data.frame(package = c("foreach", "codetools", "iterators"),
                        version = c("1.4.0", "0.2-7", "1.0.5"),
                        stringsAsFactors = FALSE)
  addOldPackage(pkgList, path = pth, repos = mirror, vers = oldVers$version, type = "source")
  # NOTE: older binary versions would need to be build from source
  
  # Check if updated packages are available
  oldPackages(path = pth, repos = mirror, type = "source") # should need update
  oldPackages(path = pth, repos = mirror, type = "win.binary") # should be current
  
  # Update available packages
  updatePackages(path = pth, repos = mirror, type = "source", ask = FALSE) # should need update
  updatePackages(path = pth, repos = mirror, type = "win.binary") # should be current
  
  # Delete temporary folder
  unlink(pth, recursive = TRUE)
}

if (interactive()) {
  availPkgs <- cranJuly2014
  
  availPkgs <- pkgAvail(
    repos = c(CRAN = "https://cloud.r-project.org"),
    type = "source"
  )
  
  
  # Create dependency graph using stored database of available packages
  p <- makeDepGraph(
    c("ggplot2", "forecast"),
    availPkgs = availPkgs
  )
  
  if(require(igraph)) plot(p)

  # Create dependency graph using newly retrieved database from CRAN
  
  p <- makeDepGraph(
    c("ggplot2", "forecast"),
    repos = c(CRAN = getOption("minicran.mran")),
    type = "source"
  )
  if(requireNamespace("igraph", quietly = TRUE)) {
    plot(p)
  } else {
    message("install package `igraph` to view dependency graph")
  }
  
}

# Specify list of packages to download
mirror <- c(CRAN = "https://cloud.r-project.org")
pkgs <- c("foreach")

if (interactive()) {
  pdb <- cranJuly2014
  
  pdb <- pkgAvail(
    repos = c(CRAN = getOption("minicran.mran")),
    type = "source"
  )
  
  pkgList <- pkgDep(pkgs, availPkgs = pdb, repos = mirror,
                    type = "source", suggests = FALSE)
  pkgList
  
  
  # Create temporary folder for miniCRAN
  dir.create(pth <- file.path(tempdir(), "miniCRAN"))
  
  # Make repo for source and win.binary
  makeRepo(pkgList, path = pth, repos = mirror, type = "source")
  
  # List all files in miniCRAN
  list.files(pth, recursive = TRUE)
  
  # Check for available packages
  pkgAvail(repos = pth, type = "source")
  
  # Repeat process for windows binaries
  makeRepo(pkgList, path = pth, repos = mirror, type = "win.binary")
  pkgAvail(repos = pth, type = "win.binary")
  
  # Delete temporary folder
  unlink(pth, recursive = TRUE)
}

# Create package database
pdb <- cranJuly2014

if (interactive()) {
  pdb <- pkgAvail(repos = c(CRAN = "https://cloud.r-project.org"))

  # Overwrite pdb with development version of miniCRAN at github
  newpdb <- addPackageListingGithub(pdb = pdb, "andrie/miniCRAN")
  newpdb["miniCRAN", ]

  # Add package from github that's not currently on CRAN
  newpdb <- addPackageListingGithub(pdb = pdb, repo = "tidyverse/ggplot2", branch = "main")
  newpdb["ggplot2", ]

  set.seed(1)
  plot(makeDepGraph("ggplot2", availPkgs = newpdb, suggests = TRUE))
}

## ----init---------------------------------------------------------------------
library("miniCRAN")

## ----pkgdep-------------------------------------------------------------------
tags <- "chron"
pkgDep(tags, availPkgs = cranJuly2014)

## ----makeDepGraph, warning=FALSE----------------------------------------------
dg <- makeDepGraph(tags, enhances = TRUE, availPkgs = cranJuly2014)
set.seed(1)
plot(dg, legendPosition = c(-1, 1), vertex.size = 20)

## ----so-tags, warning=FALSE, fig.width=10, fig.height=10----------------------
tags <- c("ggplot2", "data.table", "plyr", "knitr", "shiny", "xts", "lattice")
pkgDep(tags, suggests = TRUE, enhances = FALSE, availPkgs = cranJuly2014)

dg <- makeDepGraph(tags, enhances = TRUE, availPkgs = cranJuly2014)
set.seed(1)
plot(dg, legendPosition = c(-1, -1), vertex.size = 10, cex = 0.7)

### `checkVersions` and `add.packages.miniCRAN` require an existing miniCRAN repo

# Specify list of packages to download
mirror <- c(CRAN = "https://cloud.r-project.org")
mirror
pkgs <- c("foreach")
pkgTypes <- c("source", "win.binary")

if (interactive()) {
  if (!is.online()) {
    message("p3m seems to be not available.  Check your internet connection.")
  } else {
    pdb <- pkgAvail(repos = mirror, type = "source")
  }
} else {
  pdb <- cranJuly2014
}


if (interactive()) {
  if (!is.online()) {
    message("p3m seems to be not available.  Check your internet connection.")
  } else {
    pkgList <- pkgDep(pkgs, availPkgs = pdb, repos = mirror, type = "source", suggests = FALSE)
    pkgList
  }
}

# Create temporary folder for miniCRAN

if (interactive()) {
  if (!is.online()) {
    message("p3m seems to be not available.  Check your internet connection.")
  } else {
    dir.create(pth <- file.path(tempdir(), "miniCRAN"))
    
    # Make repo for source and win.binary
    makeRepo(pkgList, path = pth, repos = mirror, type = pkgTypes)
    
    # Add other versions of a package (and assume these were added previously)
    oldVers <- data.frame(
      package = c("foreach", "codetools", "iterators"),
      version = c("1.4.0", "0.2-7", "1.0.5"),
      stringsAsFactors = FALSE
    )
    pkgs <- oldVers$package
    addOldPackage(pkgs, path = pth, vers = oldVers$version, repos = mirror, type = "source")
    # NOTE: older binary versions would need to be build from source
    
    # List package versions in the miniCRAN repo (produces warning about duplicates)
    pkgVersionsSrc <- checkVersions(pkgs, path = pth, type = "source")
    pkgVersionsBin <- checkVersions(pkgs, path = pth, type = "win.binary")
    
    # After inspecting package versions, remove old versions
    basename(pkgVersionsSrc$source) # "foreach_1.4.0.tar.gz"  "foreach_1.4.2.tar.gz"
    basename(pkgVersionsBin$win.binary) # "foreach_1.4.0.zip"     "foreach_1.4.2.zip"
    file.remove(c(pkgVersionsSrc$source[1], pkgVersionsBin$win.binary[1]))
    
    # Rebuild package index after adding/removing files
    updateRepoIndex(pth, type = pkgTypes, Rversion = R.version)
    
    pkgAvail(pth, type = "source")
    
    # Add new packages (from CRAN) to the miniCRAN repo
    addPackage("Matrix", path = pth, repos = mirror, type = pkgTypes)
    
    # Delete temporary folder
    unlink(pth, recursive = TRUE)
  }
}

# helper functions for testing

hostname <- function(x) {
  y <- sub("https*://", "", x)
  sub("latest$", "", y)
}


# Interrupt the test if url can not be reached
skip_if_offline <- function(url = p3m()) {
  # browser()
  mran_host <- hostname(p3m())
  if (is.null(url)) url <- mran_host
  if (!is.online(url = url)) testthat::skip("offline")
}


set_test_types <- function() {
  types <-
    Sys.getenv(
      "minicran_test_scope",
      unset = "source, win.binary, mac.binary"
    )

  types <- gsub(" +", "", types)
  unname(strsplit(types, ",")[[1]])
}


# Use in unit tests to check if source or binary files are in repo
.checkForRepoFiles <- function(path, pkgList, prefix) {
  ptn <- "tar\\.gz|zip|tgz"
  ff <- list.files(file.path(path, prefix), recursive = TRUE, pattern = ptn)
  if (length(ff) < length(pkgList)) return(FALSE)
  ret <- sapply(pkgList, function(x) any(grepl(x, ff)))
  if (all(ret > 0)) TRUE else {
    message(ret)
    FALSE
  }
}

# These functions mock downloading of packages, for fast testing purposes only

#' @importFrom utils available.packages
mock_download_packages <- function(pkgs, destdir, available, type, ...) {
  if (missing(available) || is.null(available))
    available <- available.packages()
  downloadFileName <- function(package, version, type) {
    paste0(package, "_", version, pkgFileExt(type))
  }
  versions <- setNames(available[pkgs, "Version"], pkgs)
  downloads <- mapply(
    names(versions),
    versions,
    USE.NAMES = FALSE,
    FUN = function(p, v) {
      fn <- file.path(destdir, downloadFileName(p, v, type))
      writeLines("", fn)
      matrix(c(p, fn), ncol = 2)
    }
  )
  t(downloads)
}

mock_write_packages <- function(dir, type = "source", r_version) {
  pattern <- ".tgz$|.zip$|.tar.gz$"
  if (grepl("mac.binary", type)) type <- "mac.binary"
  ff <- list.files(dir, recursive = TRUE, full.names = TRUE, pattern = pattern)
  ffb <- basename(ff)
  pkgs <- ffb[!grepl("^PACKAGES.*", ffb)]
  np <- length(pkgs)
  pkg_names <- gsub(pattern, "", pkgs)
  db <- matrix(unlist(strsplit(pkg_names, "_")), ncol = 2, byrow = TRUE)
  colnames(db) <- c("Package", "Version")
  db

  if (np > 0L) {
    db[!is.na(db) & (db == "")] <- NA_character_
    con <- file(file.path(dir, "PACKAGES"), "wt")
    write.dcf(db, con)
    close(con)
    con <- gzfile(file.path(dir, "PACKAGES.gz"), "wt")
    write.dcf(db, con)
    close(con)
    rownames(db) <- db[, "Package"]
    r_version <- twodigitRversion(r_version)
    if (r_version >= "3.5.0") {
      saveRDS(db, file.path(dir, "PACKAGES.rds"))
    } else {
      saveRDS(db, file.path(dir, "PACKAGES.rds"), version = 2)
    }
  }
  np
}


# Create sample repo from p3m snapshot
.createSampleRepo <- function(p3m, path, pkgs, Rversion = "4.0", types) {
  if (missing(p3m)) p3m <- p3m("2024-01-02")
  if (missing(path)) path <- file.path(tempdir(), "miniCRAN", Sys.Date())
  if (missing(pkgs)) pkgs <- c("chron", "curl")
  if (missing(types)) types <- set_test_types()

  # pdb_source <- pkgAvail(repos = p3m, type = "source", Rversion = Rversion)
  # pdb_win    <- pkgAvail(repos = p3m, type = "win.binary", Rversion = Rversion)
  # pdb_mac    <- pkgAvail(repos = p3m, type = "mac.binary", Rversion = Rversion)

  mockr::with_mock(
    download_packages = mock_download_packages,
    write_packages = mock_write_packages,
    .env = "miniCRAN",
    {
      for (type in types) {
        pdb <- pkgAvail(repos = p3m, type = type, Rversion = Rversion)
        pkgList_source <- pkgDep(
          pkgs,
          availPkgs = pdb,
          repos = p3m,
          type = type,
          suggests = FALSE,
          Rversion = Rversion
        )

        makeRepo(
          pkgList_source,
          path = path,
          repos = p3m,
          type = type,
          quiet = TRUE,
          Rversion = Rversion
        )
      }
    }
  )
}

make_fake_package <- function(version = "0.1.0", base_path = tempdir()) {
  fake_package <- file.path(base_path, "fake.package")
  dir.create(fake_package, showWarnings = FALSE)

  # Create a fake function to add to the package
  foo <- function(x) NA

  # Create the skeleton

  # browser()

  if (getRversion() >= "3.5") {
    suppressMessages(
      package.skeleton(
        "fake.package",
        path = base_path,
        list = "foo",
        force = TRUE,
        environment = environment(foo),
        encoding = "UTF-8"
      )
    )
  } else {
    suppressMessages(
      package.skeleton(
        "fake.package",
        path = base_path,
        list = "foo",
        force = TRUE,
        environment = environment(foo)
      )
    )
  }

  # Remove unnecessary detritus from skeleton
  file.remove(file.path(fake_package, "NAMESPACE"))
  unlink(file.path(fake_package, "data"), recursive = TRUE)
  unlink(file.path(fake_package, "man"), recursive = TRUE)
  unlink(file.path(fake_package, "Read-and-delete-me"), recursive = TRUE)

  # Write a function file with some roxygen
  writeLines(
    con = file.path(fake_package, "R", "foo.R"),
    text = "
  #' Foo.
  #' 
  #' Does nothing.
  #' @export
  #' foo <- function(x)NULL
  
  "
  )

  # Set package version
  desc <- readLines(file.path(fake_package, "DESCRIPTION"))
  version_line <- grep("^Version:", desc)
  desc[version_line] <- paste0("Version: ", version)
  writeLines(desc, con = file.path(file.path(fake_package, "DESCRIPTION")))

  # Document the package

  suppressMessages(
    devtools::document(fake_package, quiet = TRUE)
  )

  # Build the package
  devtools::build(fake_package, path = base_path, quiet = TRUE)
}

{
  mirror <- p3m("2023-08-31")
  if (!is.online(mirror, tryHttp = FALSE)) {
    # Use http:// for older versions of R
    mirror <- sub("^https://", "http://", mirror)
  }
  rvers <- "4.3"
  pkgs <- c("Rcpp")
  repo_root <- file.path(tempdir(), "miniCRAN", Sys.Date())
  if (file.exists(repo_root)) unlink(repo_root, recursive = TRUE)

  # list.files(repo_root, recursive = TRUE)

  unlink(list.files(tempdir(), pattern = ".rds$", full.names = TRUE))
}

# types are defined in env var `minicran_test_scope`
types <- set_test_types()
pkg_type <- types[1]

test_that("makeRepo downloads files and builds PACKAGES", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not_installed("mockr")
  types <- set_test_types()
  for (pkg_type in types) {
    # pdb <- pkgAvail(repos = mirror, type = pkg_type, quiet = TRUE)
    pdb <- pkgAvail(
      repos = mirror,
      type = pkg_type,
      Rversion = rvers,
      quiet = TRUE
    )
    pkgList <- pkgDep(
      pkgs,
      availPkgs = pdb,
      repos = mirror,
      type = pkg_type,
      suggests = FALSE,
      Rversion = rvers,
      quiet = FALSE
    )
    prefix <- repoPrefix(pkg_type, Rversion = rvers)
    dir.create(repo_root, recursive = TRUE, showWarnings = FALSE)

    ret <- mockr::with_mock(
      download_packages = mock_download_packages,
      write_packages = mock_write_packages,
      .env = "miniCRAN",
      {
        makeRepo(
          pkgList,
          path = repo_root,
          repos = mirror,
          type = pkg_type,
          quiet = TRUE,
          Rversion = rvers
        )
      }
    )

    expect_type(ret, "character")
    expect_equal(length(ret), length(pkgList))

    expect_true(
      .checkForRepoFiles(repo_root, pkgList, prefix)
    )
    expect_true(
      file.exists(file.path(repo_root, prefix, "PACKAGES.gz"))
    )
    expect_true(
      all(
        pkgList %in%
          pkgAvail(repos = repo_root, type = pkg_type, Rversion = rvers)[,
            "Package"
          ]
      )
    )
  }
})


# types are defined in env var `minicran_test_scope`

unlink(repo_root, recursive = TRUE)

# Create a helper function to make a fake package in tempdir()

# Create fake.package

list_fake_pkg <- function(path = tempdir()) {
  list.files(path = path, pattern = "fake.package_.*.tar.gz", full.names = TRUE)
}


# tests start -------------------------------------------------------------
repo_path <- file.path(tempdir(), "cran")

test_that("can add fake local package", {
  skip_on_cran()
  skip_if_offline()

  # create the fake package
  fake_path <- file.path(tempdir(), "fake.package")
  unlink(fake_path, recursive = TRUE)
  unlink(list_fake_pkg())

  make_fake_package("0.1.0")

  expect_true(length(list_fake_pkg()) == 1)

  # Create a miniCRAN repo
  unlink(repo_path, recursive = TRUE)
  dir.create(repo_path, showWarnings = FALSE)
  repo <- "https://cran.r-project.org"
  miniCRAN::makeRepo(
    pkgs = c(),
    path = repo_path,
    type = "win.binary",
    repos = repo,
    quiet = TRUE
  )

  expect_message(
    addLocalPackage(
      pkgs = "fake.package",
      pkgPath = tempdir(),
      path = repo_path
    ),
    "copying fake.package_0.1.0.tar.gz"
  )
  expect_message(
    addLocalPackage(
      pkgs = "fake.package",
      pkgPath = tempdir(),
      path = repo_path
    ),
    "All packages up to date. Nothing to add."
  )
})


test_that("can update fake local package", {
  skip_on_cran()

  # Update fake.package

  make_fake_package("0.2.0")
  expect_message(
    addLocalPackage(
      pkgs = "fake.package",
      pkgPath = tempdir(),
      path = repo_path
    ),
    "copying fake.package_0.2.0.tar.gz"
  )
})

# make baseline repo ------------------------------------------------------

{
  repo_root <- file.path(tempdir(), "miniCRAN", Sys.Date())
  if (file.exists(repo_root)) unlink(repo_root, recursive = TRUE)
  dir.create(repo_root, recursive = TRUE, showWarnings = FALSE)

  mirror <- p3m("2023-08-31")
  if (!is.online(mirror, tryHttp = FALSE)) {
    # Use http:// for older versions of R
    mirror <- sub("^https://", "http://", mirror)
  }
  rvers <- "4.2"
  pkgs <- c("chron", "data.table")

  types <- intersect(
    set_test_types(),
    # c("source", "win.binary")
    c("source", "win.binary", "mac.binary")
  )

  names(types) <- types
  pdb <- list()
  pkgList <- list()
}

test_that("sample repo is setup correctly", {
  skip_if_offline(mirror)

  type <- types[1]
  pdb <<- lapply(
    types,
    pkgAvail,
    repos = mirror,
    Rversion = rvers,
    quiet = TRUE
  )
  expect_type(pdb, "list")
  pkgList <<- lapply(types, function(type) {
    # type <- names(type)
    pkgDep(
      pkg = pkgs,
      type = names(type),
      availPkgs = pdb[[type]],
      repos = mirror,
      suggests = FALSE,
      Rversion = rvers
    )
  })
  expect_type(pkgList, "list")

  .createSampleRepo(
    path = repo_root,
    p3m = mirror,
    Rversion = rvers,
    pkgs = pkgs,
    types = names(types)
  )
  # expect_type(z, "character")
  pkg_names <- unname(pkgAvail(
    repo_root,
    quiet = TRUE,
    type = type,
    Rversion = rvers
  )[, "Package"])
  expect_true(all(pkgs %in% pkg_names))
})


# Add packages to repo ----------------------------------------------------

pkgsAdd <- c("Rcpp")
pkg_type <- names(types)[1]

for (pkg_type in names(types)) {
  skip_if_not_installed("mockr")

  test_that(
    sprintf(
      "addPackage downloads %s files and rebuilds PACKAGES file",
      pkg_type
    ),
    {
      skip_on_cran()
      skip_if_offline(mirror)

      pkgListAdd <- pkgDep(
        pkgsAdd,
        availPkgs = pdb[[pkg_type]],
        repos = mirror,
        type = pkg_type,
        suggests = FALSE,
        Rversion = rvers
      )
      prefix <- repoPrefix(pkg_type, Rversion = rvers)

      mockr::with_mock(
        download_packages = mock_download_packages,
        write_packages = mock_write_packages,
        .env = "miniCRAN",
        {
          addPackage(
            pkgListAdd,
            path = repo_root,
            repos = mirror,
            type = pkg_type,
            quiet = TRUE,
            Rversion = rvers
          )
        }
      )

      expect_true(
        .checkForRepoFiles(repo_root, pkgListAdd, prefix)
      )
      expect_true(
        file.exists(file.path(repo_root, prefix, "PACKAGES.gz"))
      )
      expect_true(
        all(
          pkgListAdd %in%
            pkgAvail(
              repo_root,
              type = pkg_type,
              Rversion = rvers,
              quiet = TRUE
            )[, "Package"]
        )
      )
    }
  )
}


# Add local packages to repo ----------------------------------------------

pkgsAddLocal <- c("nnet")

for (pkg_type in names(types)) {
  skip_if_not_installed("mockr")
  test_that(
    sprintf("addLocalPackage copies %s files and rebuilds PACKAGES", pkg_type),
    {
      skip_on_cran()
      skip_if_offline(mirror)

      tmpdir <- file.path(tempdir(), "miniCRAN", "local", pkg_type)
      expect_true(dir.create(tmpdir, recursive = TRUE, showWarnings = FALSE))

      tmpdir <- normalizePath(tmpdir)
      expect_true(dir.exists(tmpdir))
      on.exit(unlink(tmpdir, recursive = TRUE), add = TRUE)

      # get most recent version
      mockr::with_mock(
        download_packages = mock_download_packages,
        write_packages = mock_write_packages,
        .env = "miniCRAN",
        {
          res <- download_packages(
            pkgsAddLocal,
            destdir = tmpdir,
            type = pkg_type,
            available = pkgAvail(mirror, pkg_type, rvers),
            contriburl = contribUrl(mirror, pkg_type, rvers),
            quiet = TRUE
          )
        }
      )

      # simulate older version also present in pkgPath directory
      f <- res[, 2]
      expect_true(
        file.copy(from = f, to = file.path(tmpdir, "MASS_7.3-0.tar.gz"))
      )
      expect_equal(length(list.files(tmpdir)), 2)

      mockr::with_mock(
        download_packages = mock_download_packages,
        write_packages = mock_write_packages,
        .env = "miniCRAN",
        {
          addLocalPackage(
            pkgs = pkgsAddLocal,
            pkgPath = tmpdir,
            path = repo_root,
            type = pkg_type,
            quiet = TRUE,
            Rversion = rvers
          )
        }
      )

      prefix <- repoPrefix(pkg_type, Rversion = rvers)
      expect_true(
        .checkForRepoFiles(repo_root, pkgsAddLocal, prefix)
      )
      expect_true(
        file.exists(file.path(repo_root, prefix, "PACKAGES.gz"))
      )
      expect_true(
        all(
          pkgsAddLocal %in%
            pkgAvail(repo_root, type = pkg_type, Rversion = rvers)[, "Package"]
        )
      )
    }
  )
}


# Check for updates -------------------------------------------------------

p3m_mirror <- p3m("2024-01-02")
if (!is.online(p3m_mirror, tryHttp = FALSE)) {
  # Use http:// for older versions of R
  p3m_mirror <- sub("^https://", "http://", mirror)
}

pkg_type <- names(types)[1]

for (pkg_type in names(types)) {
  test_that(
    sprintf("updatePackages downloads %s files and builds PACKAGES", pkg_type),
    {
      skip_on_cran()
      skip_if_offline(p3m_mirror)

      prefix <- repoPrefix(pkg_type, Rversion = rvers)

      old <- suppressWarnings(
        oldPackages(
          path = repo_root,
          repos = p3m_mirror,
          type = pkg_type,
          Rversion = rvers,
          quiet = FALSE
        )
      )

      # In the following allow for differences between mac.binary and other types
      expect_gte(nrow(old), 1)
      expect_equal(ncol(old), 4)
      expect_true("data.table" %in% rownames(old))

      mockr::with_mock(
        download_packages = mock_download_packages,
        write_packages = mock_write_packages,
        .env = "miniCRAN",
        {
          updatePackages(
            path = repo_root,
            repos = p3m_mirror,
            type = pkg_type,
            ask = FALSE,
            quiet = TRUE,
            Rversion = rvers
          )
        }
      )

      updateVers <- getPkgVersFromFile(
        list.files(file.path(repo_root, prefix))
      )

      expect_true(
        .checkForRepoFiles(repo_root, pkgList[[pkg_type]], prefix)
      )

      expect_true(
        file.exists(file.path(repo_root, prefix, "PACKAGES.gz"))
      )

      mockr::with_mock(
        download_packages = mock_download_packages,
        write_packages = mock_write_packages,
        .env = "miniCRAN",
        {
          old <- oldPackages(
            path = repo_root,
            repos = p3m_mirror,
            type = pkg_type,
            Rversion = rvers
          )
        }
      )
      # browser()
      expect_equal(nrow(old), 0)
      expect_equal(ncol(old), 4)
    }
  )
}


# Check for duplicate packages --------------------------------------------

pkg_type <- names(types)[3]
for (pkg_type in names(types)) {
  test_that(
    sprintf("checkVersions() finds out-of-date %s packages", pkg_type),
    {
      skip_on_cran()
      skip_if_offline(p3m_mirror)

      oldVersions <- list(package = c("acepack"), version = c("1.3-2"))

      if (pkg_type != "source") {
        expect_error(
          mockr::with_mock(
            download_packages = mock_download_packages,
            write_packages = mock_write_packages,
            .env = "miniCRAN",
            {
              addOldPackage(
                oldVersions[["package"]],
                path = repo_root,
                vers = oldVersions[["version"]],
                repos = p3m_mirror,
                type = pkg_type
              )
            }
          )
        )
      } else {
        mockr::with_mock(
          download_packages = mock_download_packages,
          write_packages = mock_write_packages,
          .env = "miniCRAN",
          {
            addOldPackage(
              oldVersions[["package"]],
              path = repo_root,
              vers = oldVersions[["version"]],
              repos = p3m_mirror,
              type = pkg_type
            )
          }
        )
        files <- suppressWarnings(
          checkVersions(path = repo_root, type = pkg_type)[[pkg_type]]
        )

        expect_true(
          all(file.exists(files))
        )

        pkgs <- sapply(strsplit(basename(files), "_"), "[[", 1)
        dupes <- pkgs[duplicated(pkgs)]
        expect_true(
          all(dupes == oldVersions[["package"]])
        )
      }
    }
  )
}

checkPkgDepFunctions <- function(
  pkg,
  availPkgs = cranJuly2014,
  repos = p3m(),
  type = "source",
  suggests = TRUE,
  enhances = FALSE,
  includeBasePkgs = FALSE
) {
  p1 <- pkgDep(
    pkg,
    availPkgs = availPkgs,
    repos = repos,
    type = type,
    suggests = suggests,
    enhances = enhances,
    includeBasePkgs = includeBasePkgs
  )
  p2 <- makeDepGraph(
    pkg,
    availPkgs = availPkgs,
    repos = repos,
    type = type,
    suggests = suggests,
    enhances = enhances,
    includeBasePkgs = includeBasePkgs
  )

  vnames <- igraph::V(p2)$name
  diff1 <- setdiff(vnames, p1)
  diff2 <- setdiff(p1, vnames)
  result <- length(diff1) == 0 & length(diff2) == 0
  if (!result) {
    msg <- paste0(
      "\nmakeDepGraph() results not in pkgDep(): \n - ",
      paste(diff1, collapse = ", "),
      "\npkgDep() results not in makeDepGraph(): \n - ",
      paste(diff2, collapse = ", ")
    )

    warning(msg)
  }
  result
}


mock_require <- function(pkg, ...) {
  packages.to.exclude <- c("igraph")
  inSearchPath <- any(
    grepl(
      sprintf("package:%s$", paste(packages.to.exclude, collapse = "|")),
      search()
    )
  )
  if (inSearchPath) stop("Required package already in search path")

  package <- as.character(substitute(pkg))
  if (package %in% packages.to.exclude) FALSE else
    base::requireNamespace(package, character.only = TRUE, ...)
}


test_that("plots depGraph", {
  tag <- "MASS"
  g <- makeDepGraph(tag, availPkgs = cranJuly2014)
  pdf(tempfile(fileext = ".pdf"))
  expect_null(
    plot.pkgDepGraph(g, legendPosition = NULL)
  )
  dev.off()
})

test_that("makeDepGraph and pgkDep gives similar results for MASS", {
  skip_if_offline()

  tag <- "MASS"

  expect_true(
    checkPkgDepFunctions(tag)
  )

  skip_on_cran()

  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE)
  )

  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, suggests = FALSE)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, enhances = TRUE)
  )
})


test_that("makeDepGraph and pgkDep gives similar results for chron", {
  skip_on_cran()

  tag <- "chron"

  expect_true(
    checkPkgDepFunctions(tag)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, suggests = FALSE)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, enhances = TRUE)
  )
})


test_that("makeDepGraph and pgkDep gives similar results for data.table", {
  skip_on_cran()

  tag <- "data.table"

  expect_true(
    checkPkgDepFunctions(tag)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, suggests = FALSE)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, enhances = TRUE)
  )
})

test_that("makeDepGraph and pgkDep gives similar results for ggplot2", {
  skip_on_cran()

  tag <- "ggplot2"

  expect_true(
    checkPkgDepFunctions(tag)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, suggests = FALSE)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, enhances = TRUE)
  )
})


test_that("makeDepGraph and pgkDep gives similar results for complex query", {
  skip_on_cran()

  tag <- c("ggplot2", "data.table", "plyr", "knitr", "shiny", "xts", "lattice")

  expect_true(
    checkPkgDepFunctions(tag)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, suggests = FALSE)
  )
  expect_true(
    checkPkgDepFunctions(tag, includeBasePkgs = TRUE, enhances = TRUE)
  )
})

test_that("can read CRAN description", {
  skip_on_cran()
  skip_if_offline()
  skip_if(getRversion() <= "3.4")

  skip_if_not_installed("mockery")
  skip_if_not_installed("testthis")

  mockery::stub(
    getCranDescription,
    what = "tools::CRAN_package_db",
    function(...) {
      suppressMessages(
        testthis::read_testdata("/pdb.rds")
      )
    }
  )
  p <- getCranDescription("ggplot2", repos = p3m("2024-01-02"))
  expect_s3_class(p, "data.frame")
  expect_equal(p$Package[1], "ggplot2")
})

test_that("can add package from github", {
  skip_on_cran()
  skip_if_offline()

  pdb <- pkgAvail(repos = p3m("2024-01-02"))
  expect_type(pdb, "character")

  # Overwrite pdb with development version of miniCRAN at github
  expect_warning(
    newpdb <- addPackageListingGithub(pdb = pdb, "andrie/miniCRAN"),
    "Overwriting package information for: miniCRAN"
  )
  expect_type(newpdb, "character")
  expect_equal(nrow(pdb), nrow(newpdb))
  expect_equal(ncol(pdb), ncol(newpdb))
})

test_that("pkgAvail throws warnings and errors for incorrect CRAN repos", {
  expect_warning(pkgAvail(repos = ""), "unable to access index for repository")

  is.available.packages <- function(x) {
    required_cols <- c("Package", "Version", "Depends", "Imports", "Suggests")
    all(
      is.matrix(x),
      nrow(x) > 0,
      required_cols %in% colnames(x)
    )
  }

  skip_if_offline(p3m())

  mran <- p3m("2024-01-02")

  expect_true(is.available.packages(
    pkgAvail(repos = mran)
  ))
})

if (interactive()) {
  library(testthat)
  Sys.setenv(NOT_CRAN = "true")
}


{
  mirror <- p3m("2024-01-02")
  if (!is.online(mirror, tryHttp = FALSE)) {
    # Use http:// for older versions of R
    mirror <- sub("^https://", "https://", mirror)
  }
  rvers <- "4.0"
  pkgs <- c("MASS")
  repo_root <- file.path(tempdir(), "miniCRAN", Sys.Date())
  new_repo_root <- file.path(tempdir(), "newMiniCRAN", Sys.Date())
  if (file.exists(repo_root)) unlink(repo_root, recursive = TRUE)
  if (file.exists(new_repo_root)) unlink(new_repo_root, recursive = TRUE)

  # list.files(repo_root, recursive = TRUE)
}

types <- c("win.binary")
names(types) <- c("win.binary")

pkg_type <- names(types)

test_that("makeRepo downloads files and builds PACKAGES file", {
  skip_if_offline(mirror)
  # skip_on_cran()
  for (pkg_type in names(types)) {
    # Create local miniCRAN
    # mockery::stub(makeRepo, "download_packages", mock_download_packages, depth = 1)
    # mockery::stub(updateRepoIndex, "write_packages", mock_write_packages, depth = 1)

    pdb <- pkgAvail(
      repos = mirror,
      type = pkg_type,
      Rversion = rvers,
      quiet = TRUE
    )
    pkgList <- pkgDep(
      pkgs,
      availPkgs = pdb,
      repos = mirror,
      type = pkg_type,
      suggests = FALSE,
      Rversion = rvers
    )
    prefix <- repoPrefix(pkg_type, Rversion = rvers)
    dir.create(repo_root, recursive = TRUE, showWarnings = FALSE)

    ret <- makeRepo(
      pkgList,
      path = repo_root,
      repos = mirror,
      type = pkg_type,
      quiet = TRUE,
      Rversion = rvers
    )

    expect_type(ret, "character")
    expect_equal(length(ret), length(pkgList))

    expect_true(
      .checkForRepoFiles(repo_root, pkgList, prefix)
    )
    expect_true(
      file.exists(file.path(repo_root, prefix, "PACKAGES.gz"))
    )
    expect_true(
      all(
        pkgList %in%
          pkgAvail(repos = repo_root, type = pkg_type, Rversion = rvers)[,
            "Package"
          ]
      )
    )

    # Create miniCRAN from existing miniCRAN

    localCRAN <- paste0("file:///", repo_root)

    pdb <- pkgAvail(repos = localCRAN, type = pkg_type, Rversion = rvers)
    pkgList <- pkgDep(
      pkgs,
      availPkgs = pdb,
      repos = localCRAN,
      type = pkg_type,
      suggests = FALSE,
      Rversion = rvers
    )
    prefix <- repoPrefix(pkg_type, Rversion = rvers)
    dir.create(new_repo_root, recursive = TRUE, showWarnings = FALSE)

    ret <- makeRepo(
      pkgList,
      path = new_repo_root,
      repos = localCRAN,
      type = pkg_type,
      quiet = TRUE,
      Rversion = rvers
    )

    expect_type(ret, "character")
    expect_equal(length(ret), length(pkgList))

    expect_true(
      .checkForRepoFiles(new_repo_root, pkgList, prefix)
    )
    expect_true(
      file.exists(file.path(new_repo_root, prefix, "PACKAGES.gz"))
    )
    expect_true(
      all(
        pkgList %in%
          pkgAvail(repos = new_repo_root, type = pkg_type, Rversion = rvers)[,
            "Package"
          ]
      )
    )
  }
})

unlink(repo_root, recursive = TRUE)

.onLoad <- function(libname, pkgname) {
  mran.url <- if (getRversion() >= "3.2.2") {
    "https://packagemanager.posit.co/cran" ## use HTTPS
  } else {
    "https://packagemanager.posit.co/cran" ## use HTTP
  }

  ## set options using the approach used by devtools
  opts <- options()
  opts.miniCRAN <- list(
    minicran.mran = mran.url
  )
  toset <- !(names(opts.miniCRAN) %in% names(opts))
  if (any(toset)) options(opts.miniCRAN[toset])

  invisible()
}

p3m <- function(snapshot = NULL) {
  url <- getOption("minicran.mran")
  if (missing("snapshot") || is.null(snapshot)) {
    url
  } else {
    sprintf("%s/%s", url, snapshot)
  }
}

CRAN <- function() getOption("repos")[1]


#' Returns TRUE if the p3m URL can be accessed.
#'
#' @param url p3m url
#' @param tryHttp If TRUE, also attempts http URL, for compatibility with older versions of R
#'
#' @export
is.online <- function(url = NULL, tryHttp = TRUE) {
  if (is.null(url)) url <- p3m()
  url <- sub("latest$", "", url)
  url <- sub("\\d{4}-\\d{2}-\\d{2}$", "", url)

  readFromUrl <- function(url) {
    z <- tryCatch(
      suppressWarnings(readLines(url, n = 1, warn = FALSE)),
      error = function(e) e
    )
    !inherits(z, "error")
  }

  if (!readFromUrl(url)) {
    if (grepl("^https://", url) && tryHttp) {
      url <- sub("^https://", "http://", url) # for older versions of R
      if (!readFromUrl(url)) return(FALSE)
    } else {
      return(FALSE)
    }
  }

  TRUE
}

test_that("is online", {
  skip_on_cran()

  # Should always be TRUE, but will trigger github action failure if
  # p3m is offline
  expect_true(is.online())
  expect_true(is.online(p3m("2024-01-02")))
})

#' @importFrom assertthat assert_that on_failure<-

# path --------------------------------------------------------------------

is_path <- function(x) {
  is.character(x) && length(x) == 1
}

on_failure(is_path) <- function(call, env) {
  paste0(deparse(call$x), " is not a valid path")
}

path_exists <- function(x) {
  file.exists(x)
}

on_failure(path_exists) <- function(call, env) {
  paste0("Download path ", deparse(call$x), " does not exist")
}


# pkg ---------------------------------------------------------------------

is_package <- function(x) {
  is.character(x) && length(x) > 0
}

on_failure(is_package) <- function(call, env) {
  paste0(deparse(call$x), " should be a character vector with package names")
}

is_package_vector <- is_package


# repos -------------------------------------------------------------------

is_repos <- function(x) {
  is.character(x) && length(x) > 0
}

on_failure(is_repos) <- function(call, env) {
  paste0(deparse(call$x), " should be a character vector")
}

#' Downloads packages from CRAN to specified path and creates a local
#' repository.
#'
#' Given a list of packages, downloads these packages to a specified destination
#' folder using the required CRAN folder structure, and finally creates the
#' PACKAGES index file.  Since the folder structure mimics the required
#' structure and files of a CRAN repository, it supports functions like
#' [utils::install.packages()].
#'
#' @template repo_folder_structure
#'
#' @note Internally makes use of [utils::download.packages()] and
#'   [write_PACKAGES()]
#'
#' @inheritParams pkgDep
#'
#' @inheritParams pkgAvail
#'
#' @param pkgs Character vector of packages to download
#'
#' @param path Destination download path. This path is the root folder of your
#'   new repository.
#'
#' @template Rversion
#'
#' @param download If TRUE downloads packages.
#'
#' @param quiet If TRUE, suppress status messages (if any), and the progress bar
#'   during download.
#'
#' @param writePACKAGES If TRUE, calls [write_PACKAGES()] to update the
#'   repository PACKAGES file.
#'
#' @export
#'
#' @return character vector of downloaded package files
#'
#' @family update repo functions
#'
#' @importFrom utils download.packages
#'
#' @example /inst/examples/example_makeRepo.R
makeRepo <- function(
  pkgs,
  path,
  repos = getOption("repos"),
  type = "source",
  Rversion = R.version,
  download = TRUE,
  writePACKAGES = TRUE,
  filters = NULL,
  quiet = FALSE
) {
  assert_that(is_path(path))
  # if (!file.exists(path)) stop("Download path does not exist")
  assert_that(path_exists(path))

  assert_that(is_repos(repos))

  downloaded <- lapply(type, function(t) {
    pkgPath <- repoBinPath(path = path, type = t, Rversion = Rversion)
    if (!file.exists(pkgPath)) {
      result <- dir.create(pkgPath, recursive = TRUE, showWarnings = FALSE)
      if (result) {
        if (!quiet) message("Created new folder: ", pkgPath)
      } else {
        stop("Unable to create repo path: ", pkgPath)
      }
    }

    pdb <- pkgAvail(
      repos = repos,
      type = t,
      Rversion = Rversion,
      filters = filters
    )

    if (download) {
      download_packages(
        pkgs,
        destdir = pkgPath,
        available = pdb,
        repos = repos,
        contriburl = contribUrl(repos, t, Rversion),
        type = t,
        quiet = quiet
      )
    }
  })

  if (download) {
    downloaded <- downloaded[[1]][, 2]

    ## allow for more than one repo
    fromLocalRepos <- grepl("^file://", repos)

    if (any(fromLocalRepos)) {
      # need to copy files to correct folder
      if (sum(fromLocalRepos) > 1)
        warning(
          "More than one local repos provided. Only the first listed will be used."
        )
      pat <- ifelse(Sys.info()["sysname"] == "Windows", "^file:///", "^file://")
      repoPath <- gsub(pat, "", repos[fromLocalRepos][1])
      repoPath <- normalizePath(repoPath, winslash = "/")
      path <- normalizePath(path, winslash = "/")
      downloaded <- normalizePath(downloaded, winslash = "/")
      newPath <- gsub(repoPath, path, downloaded)
      file.copy(downloaded, newPath)
      downloaded <- newPath
    }
  }

  if (writePACKAGES)
    updateRepoIndex(path = path, type = type, Rversion = Rversion)
  if (download) downloaded else character(0)
}


#' @rdname makeRepo
#' @export
updateRepoIndex <- function(path, type = "source", Rversion = R.version) {
  n <- lapply(type, function(t) {
    pkgPath <- repoBinPath(path = path, type = t, Rversion = Rversion)
    if (grepl("mac.binary", t)) t <- "mac.binary"
    # browser()
    write_packages(dir = pkgPath, type = t, r_version = Rversion)
    if (twodigitRversion(Rversion) < "3.5.0") {
      pfiles <- list.files(
        path = pkgPath,
        pattern = "PACKAGES.rds",
        recursive = TRUE,
        full.names = TRUE
      )
      for (f in pfiles) {
        pf <- readRDS(f)
        saveRDS(pf, file = f, version = 2)
      }
    }
  })
  names(n) <- type
  n
}


#' Deprecated function to download packages to local folder.
#'
#' @inheritParams makeRepo
#' @export
makeLibrary <- function(pkgs, path, type = "source") {
  .Deprecated("makeRepo")
  NULL
}

# Thin wrapper around utils::download.packages
download_packages <- function(
  pkgs,
  destdir,
  available,
  repos,
  contriburl,
  type,
  quiet
) {
  utils::download.packages(
    pkgs = pkgs,
    destdir = destdir,
    available = available,
    repos = repos,
    contriburl = contriburl,
    type = type,
    quiet = quiet
  )
}

# Thin wrapper around write_PACKAGES
write_packages <- function(dir, type, r_version) {
  r_version <- twodigitRversion(r_version)
  tools::write_PACKAGES(dir = dir, type = type)
}

# Code copied from the pkgDepTools project
# Copyright (C) Seth Falcon
# https://www.bioconductor.org/packages/release/bioc/html/pkgDepTools.html
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation

# Code copied from the pkgDepTools project
# Copyright (C) Seth Falcon
# https://www.bioconductor.org/packages/release/bioc/html/pkgDepTools.html

# Copy of tools:::split_op_version.

# @rdname pkgDepTools
# @keywords internal
split_op_version <- function(x) {
  pat <- "^([^\\([:space:]]+)[[:space:]]*\\(([^\\)]+)\\).*"
  x1 <- sub(pat, "\\1", x)
  x2 <- sub(pat, "\\2", x)
  if (x2 != x1) {
    pat <- "[[:space:]]*([[<>=!]+)[[:space:]]+(.*)"
    version <- sub(pat, "\\2", x2)
    if (!grepl("^r", version)) version <- package_version(version)
    list(name = x1, op = sub(pat, "\\1", x2), version = version)
  } else list(name = x1)
}


# Copy of tools:::.split_dependencies.

# @rdname pkgDepTools
# @keywords internal
split_dependencies <- function(x) {
  if (!length(x)) return(list())
  x <- unlist(strsplit(x, ","))
  x <- sub("[[:space:]]+$", "", x)
  x <- unique(sub("^[[:space:]]*(.*)", "\\1", x))
  names(x) <- sub("^([[:alnum:].]+).*$", "\\1", x)
  lapply(x, split_op_version)
}


# Clean package fields.
#
# Given the value from a field like 'Depends' in a package's DESCRIPTION file,
# return a character vector of package names with the version restrictions
# stripped and \R~removed.

# @param val Value from a field like 'Depends' in a package's DESCRIPTION file
# @rdname pkgDepTools
# @keywords internal
cleanPkgField <- function(val) {
  if (is.na(val)) return(character(0))
  val <- names(split_dependencies(val))
  if (is.null(val)) return(character(0))
  val <- val[!val %in% "R"]
  if (length(val)) return(val)
  character(0)
}

#' Plots a package dependency graph.
#'
#' @param x Object to plot
#' @param pkgsToHighlight Optional character vector with names of package to
#'   highlight. If missing, defaults to packages used in original call to
#'   [makeDepGraph()]
#' @param main Title of plot
#' @param legendPosition Numeric vector of length 2, indicating (x, y) position
#'   of edge legend. Both values should be in the range `[-1; 1]`.  If `NULL`,
#'   the edge legend is not displayed.
#' @param shape Shape of edge.  See [igraph::igraph.plotting()]. Could be
#'   "none", "circle", "square", ...
#' @param vertex.size Size of vertex shape. [igraph::igraph.plotting()]
#' @param cex Vertex label size.
#' @param ... Ignored
#'
#' @importFrom igraph V plot.igraph
#' @export
#'
#' @family dependency functions
#'
#' @example inst/examples/example_plot.pkgDepGraph.R
#'
plot.pkgDepGraph <- function(
  x,
  pkgsToHighlight,
  main = paste(attr(x, "pkgs"), collapse = ", "),
  legendPosition = c(-1.2, -1),
  shape = "circle",
  vertex.size = 8,
  cex = 1,
  ...
) {
  class(x) <- "igraph"
  plotColours <- c("grey80", "orange")
  if (missing("pkgsToHighlight")) {
    pkgsToHighlight <- attr(x, "pkgs")
  }

  topLevel <- as.numeric(igraph::V(x)$name %in% pkgsToHighlight)
  vColor <- plotColours[1 + topLevel]
  vFont <- 1 + topLevel
  vShape <- c("none", shape)[1 + topLevel]

  edgeColor <- c(
    Imports = "red",
    Depends = "orange",
    Suggests = "grey80",
    Enhances = "blue",
    LinkingTo = "black"
  )
  eColor <- edgeColor[igraph::edge_attr(x, "type")]

  typesInGraph <- unique(igraph::edge_attr(x, "type"))
  edgeColor <- edgeColor[typesInGraph]
  # eColor    <- eColor[typesInGraph]

  par(mai = rep(0.25, 4))
  # browser()

  igraph::plot.igraph(
    x,
    vertex.size = vertex.size,
    edge.arrow.size = 0.5,
    edge.color = eColor,
    vertex.label.cex = cex,
    vertex.label.color = "black",
    vertex.color = vColor,
    vertex.shape = vShape,
    vertex.label.font = vFont,
    xlim = c(-1.5, 1)
  )
  pch1 <- rep(19, length(plotColours))
  pch2 <- rep(-8594, length(edgeColor)) #https://www.alanwood.net/unicode/arrows.html
  yjust <- function(x) 0.5 * (x + 1)
  xjust <- function(x) 1

  # Edge legend
  if (!is.null(legendPosition)) {
    legend(
      x = legendPosition[1],
      y = legendPosition[2],
      xjust = xjust(legendPosition[1]),
      yjust = yjust(legendPosition[2]),
      legend = names(edgeColor),
      col = edgeColor,
      pch = pch2,
      y.intersp = 0.75,
      cex = cex
    )
  }
  title(main, cex = cex)
}

fastdf <- function(list) {
  rows <- unique(unlist(lapply(list, NROW)))
  stopifnot(length(rows) == 1)
  class(list) <- "data.frame"
  attr(list, "row.names") <- c(NA_integer_, -rows)
  list
}

addDepType <- function(
  p,
  type = c("Imports", "Depends", "LinkingTo", "Suggests"),
  pdb
) {
  if (!p %in% rownames(pdb)) {
    fastdf(list(
      dep = character(0),
      package = character(0),
      type = character(0)
    ))
  } else {
    x <- cleanPkgField(pdb[p, type])
    fastdf(list(
      dep = x,
      package = rep(p, length(x)),
      type = rep(type, length(x))
    ))
  }
}


#' Create dependency graph from available packages.
#'
#' Each package is a node, and a dependency is an edge
#'
#' @inheritParams pkgDep
#' @inheritParams makeRepo
#'
#' @export
#' @family dependency functions
#'
#' @seealso [pkgDep()] to extract package dependencies
#'
#' @example /inst/examples/example_makeDepGraph.R
makeDepGraph <- function(
  pkg,
  availPkgs,
  repos = getOption("repos"),
  type = "source",
  suggests = TRUE,
  enhances = FALSE,
  includeBasePkgs = FALSE,
  ...
) {
  if (missing(availPkgs)) {
    availPkgs <- pkgAvail(repos = repos, type = type)
  }
  pkgs <- availPkgs
  rownames(pkgs) <- as.vector(pkgs[, "Package"])
  allPkgs <- rownames(pkgs)
  if (!length(allPkgs)) stop("no packages in specified repositories")

  pkgEdge <- function(p, type = c("Imports", "Depends", "LinkingTo"), pdb) {
    do.call(rbind, lapply(type, function(t) addDepType(p, t, pdb = pdb)))
  }
  pkgEdges <- function(pp, type = c("Imports", "Depends", "LinkingTo"), pdb) {
    do.call(rbind, lapply(pp, pkgEdge, type = type, pdb = pdb))
  }

  # Build depends edge list for original pkg list, combined with top level suggests

  pkg_orig <- pkg

  if (suggests) {
    edges1 <- pkgEdges(pkg, type = c("Suggests"), availPkgs)
    p_sug <- unique(unlist(
      tools::package_dependencies(
        pkg,
        db = availPkgs,
        which = "Suggests",
        recursive = FALSE
      )
    ))
    pkg <- unique(c(p_sug, pkg))
  }

  if (enhances) {
    edges2 <- pkgEdges(pkg_orig, type = c("Enhances"), availPkgs)
    p_enh <- unique(unlist(
      tools::package_dependencies(
        pkg_orig,
        db = availPkgs,
        which = "Enhances",
        recursive = FALSE
      )
    ))
    pkg <- unique(c(p_enh, pkg))
  }
  p_dep <- unique(unlist(
    tools::package_dependencies(
      pkg,
      db = availPkgs,
      which = c("Imports", "Depends", "LinkingTo"),
      recursive = TRUE
    )
  ))
  pkg <- unique(c(p_dep, pkg))

  edges <- pkgEdges(pkg, type = c("Imports", "Depends", "LinkingTo"), availPkgs)

  if (suggests) {
    edges <- rbind(edges, edges1)
  }
  if (enhances) {
    edges <- rbind(edges, edges2)
  }
  nedges <- nrow(edges)
  if (nedges && !includeBasePkgs)
    edges <- edges[!(edges[["dep"]] %in% basePkgs()), ]

  vert <- unique(c(pkg_orig, edges[["dep"]], edges[["package"]]))
  ret <- igraph::graph_from_data_frame(
    d = edges,
    directed = TRUE,
    vertices = vert
  )
  class(ret) <- c("pkgDepGraph", "igraph")
  attr(ret, "pkgs") <- pkg_orig
  ret
}

pkgFileExt <- function(type) {
  switch(
    type,
    "source" = ".tar.gz",
    "win.binary" = ".zip",
    "mac.binary" = ".tgz",
    "mac.binary.mavericks" = ".tgz",
    "mac.binary.leopard" = ".tgz",
    "mac.binary.el-capitan" = ".tgz",
    stop("Type ", type, " not recognised.")
  )
}


getPkgVersFromFile <- function(file) {
  file <- grep(
    "\\.(tar\\.gz|zip|tgz)$",
    basename(as.character(file)),
    value = TRUE
  )
  if (length(file)) {
    file <- sapply(strsplit(file, "\\.tar\\.gz"), "[[", 1)
    file <- sapply(strsplit(file, "\\.zip"), "[[", 1)
    file <- sapply(strsplit(file, "\\.tgz"), "[[", 1)
    pkg <- sapply(strsplit(file, "_"), "[[", 1)
    vers <- sapply(strsplit(file, "_"), "[[", 2)
    df <- data.frame(
      package = pkg,
      version = vers,
      stringsAsFactors = FALSE,
      row.names = NULL
    )
    df <- df[order(df$package), ]
    row.names(df) <- seq_len(nrow(df))
    df
  } else {
    data.frame(
      package = character(0),
      version = character(0),
      stringsAsFactors = FALSE
    )
  }
}


readDescription <- function(file) {
  stopifnot(file.exists(file))
  trimSpaces <- function(x) {
    gsub("^\\s+|\\s+$", "", x)
  }

  prepDescription <- function(d) {
    clean <- function(x) gsub("\n", " ", x)
    if (is.null(d$Depends)) d$Depends <- NA else d$Depends <- clean(d$Depends)
    if (is.null(d$Imports)) d$Imports <- NA else d$Imports <- clean(d$Imports)
    if (is.null(d$Suggests)) d$Suggests <- NA else
      d$Suggests <- clean(d$Suggests)
    if (is.null(d$LinkingTo)) d$LinkingTo <- NA else
      d$LinkingTo <- clean(d$LinkingTo)
    if (is.null(d$Enhances)) d$Enhances <- NA else
      d$Enhances <- clean(d$Enhances)
    d
  }
  dcf <- read.dcf(file)
  dcfList <- setNames(as.list(dcf[1, ]), colnames(dcf))
  ret <- lapply(dcfList, trimSpaces)
  prepDescription(ret)
}


addPackageListing <- function(pdb = pkgAvail(), dcf, warnings = TRUE) {
  pkgName <- dcf[["Package"]]
  pkgRow <- match(pkgName, pdb[, "Package"])

  # Build a new row matching pdb's actual column structure, filled by name
  # so it stays correct as available.packages() gains or loses columns across R versions
  newRow <- setNames(rep(NA_character_, ncol(pdb)), colnames(pdb))
  known_fields <- c(
    "Package", 
    "Version", 
    "Depends", 
    "Imports",
    "LinkingTo", 
    "Suggests", 
    "Enhances", 
    "License"
  )
  for (field in known_fields) {
    if (field %in% names(newRow) && !is.null(dcf[[field]])) {
      newRow[field] <- dcf[[field]]
    }
  }

  if (!is.na(pkgRow)) {
    pdb[pkgRow, ] <- newRow
    if (warnings) {
      warning(sprintf("Overwriting package information for: %s", pkgName))
    }
  } else {
    pdb <- rbind(pdb, newRow)
    rownames(pdb)[nrow(pdb)] <- pkgName
  }
  pdb
}


# from https://stackoverflow.com/questions/13163248 Possible to override the
# blocking of a package's (re-)installation after it has been required/loaded?

#' @importFrom httr GET stop_for_status content
readDescriptionGithub <- function(
  repo,
  username = NULL,
  branch = "master",
  quiet = TRUE
) {
  # browser()
  if (!is.null(username)) repo <- paste(username, repo, sep = "/")
  pkg <- sprintf("https://github.com/%s/raw/%s/DESCRIPTION", repo, branch)
  ff <- tempfile()
  on.exit(unlink(ff))
  request <- GET(pkg)
  stop_for_status(request)
  writeBin(content(request), ff)

  readDescription(ff)
}


#' Add DESCRIPTION information from package on github.
#'
#' Downloads the DESCRIPTION file from a package on github, parses the fields
#' and adds (or replaces) a row in the available package database.
#'
#' @param pdb Package database, usually the result of [pkgAvail()] or
#'   [available.packages()]
#' @param repo Character vector. Name of repository on github, e.g.
#'   `"andrie/rrd"`
#' @param username Optional character vector. Name of repository on github, e.g.
#'   `"andrie/rrd"`
#' @param branch name of branch, defaults to `"main"`

#' @export
#' @family github functions
#'
#' @example /inst/examples/example_addPackageListingGithub.R
addPackageListingGithub <- function(
  pdb = pkgAvail(),
  repo,
  username = NULL,
  branch = "main"
) {
  desc <- readDescriptionGithub(
    repo = repo,
    username = username,
    branch = branch
  )
  addPackageListing(pdb, desc)
}

if (interactive()) library(testthat)

# Copyright (c) Andrie de Vries 2018

test_that("package file extension is computed correctly", {
  expect_equal(pkgFileExt("source"), ".tar.gz")
  expect_equal(pkgFileExt("win.binary"), ".zip")
  expect_equal(pkgFileExt("mac.binary"), ".tgz")
  expect_error(pkgFileExt("nonsense"))
})

test_that("package verson is extracted correctly", {
  skip_on_cran()
  ff <- c(
    "KernSmooth_2.23-15.tar.gz",
    "MASS_7.3-47.tar.gz",
    "Matrix_1.2-12.tar.gz",
    "boot_1.3-20.tar.gz"
  )
  df <- data.frame(
    package = c("boot", "KernSmooth", "MASS", "Matrix"),
    version = c("1.3-20", "2.23-15", "7.3-47", "1.2-12"),
    stringsAsFactors = FALSE
  )

  # <don't understand> why we need to reorder to get tests to pass
  df <- df[order(df$package), ]
  row.names(df) <- seq_len(nrow(df))
  # </don't understand>

  expect_equal(
    getPkgVersFromFile(ff),
    df
  )

  expect_equal(
    getPkgVersFromFile("nonsense"),
    data.frame(
      package = character(0),
      version = character(0),
      stringsAsFactors = FALSE
    )
  )
})

test_that("readDescription reads file", {
  sf <- system.file("DESCRIPTION", package = "miniCRAN")
  desc <- readDescription(sf)

  expect_type(desc, "list")
  expect_true(
    all(c("Imports", "Suggests", "Package") %in% names(desc))
  )
})

#' Get the path to the repo directory containing the package files.
#'
#' @note Not all versions of R are compatible with with all package types (e.g.,
#'   `mac.binary.el-capitan` is only valid for R > 3.4.0).
#'
#' @template Rversion
#'
#' @param type  character, indicating the type of package to download and
#'   install. See [install.packages()].
#'
#' @template repo_folder_structure
#'
#' @return The file path to the package files directory.
#'
#' @keywords Internal
#'
repoPrefix <- function(type, Rversion) {
  Rversion <- twodigitRversion(Rversion)

  if (
    (type == "mac.binary.el-capitan") && (numeric_version(Rversion) < "3.4")
  ) {
    warning("Type mac.binary.el-capitan only valid for R >= 3.4")
  } else if (
    (type == "mac.binary.mavericks") && (numeric_version(Rversion) >= "3.4")
  ) {
    warning("Type mac.binary.mavericks only valid for R < 3.4")
  }

  switch(
    type,
    "source" = "src/contrib",
    "win.binary" = sprintf("bin/windows/contrib/%s", Rversion),
    "mac.binary" = sprintf("bin/macosx/contrib/%s", Rversion),
    "mac.binary.el-capitan" = sprintf(
      "bin/macosx/el-capitan/contrib/%s",
      Rversion
    ),
    "mac.binary.leopard" = sprintf("bin/macosx/leopard/contrib/%s", Rversion),
    "mac.binary.mavericks" = sprintf(
      "bin/macosx/mavericks/contrib/%s",
      Rversion
    ),
    "mac.binary.big-sur-x86_64" = sprintf(
      "bin/macosx/big-sur-x86_64/contrib/%s",
      Rversion
    ),
    "mac.binary.big-sur-arm64" = sprintf(
      "bin/macosx/big-sur-arm64/contrib/%s",
      Rversion
    ),
    stop("Type ", type, " not recognised.")
  )
}

#' Construct path to full binary location
#' @inheritParams makeRepo
#' @inheritParams repoPrefix
#'
#' @keywords Internal
#' @noRd
repoBinPath <- function(path, type, Rversion) {
  normalizePath(
    file.path(path, repoPrefix(type, Rversion)),
    mustWork = FALSE,
    winslash = "/"
  )
}

#' Get a two-digit version of the R version
#'
#' @template Rversion
#'
#' @return A character string representing the two-digit R version.
#'
#' @importFrom methods is
#'
#' @keywords Internal
#'
twodigitRversion <- function(Rversion = R.version) {
  if ("simple.list" %in% is(Rversion)) {
    paste(
      Rversion$major,
      strsplit(Rversion$minor, ".", fixed = TRUE)[[1L]][1L],
      sep = "."
    )
  } else if ("R_system_version" %in% is(Rversion)) {
    paste(
      strsplit(as.character(Rversion), ".", fixed = TRUE)[[1L]][1L:2L],
      collapse = "."
    )
  } else if (is.character(Rversion)) {
    paste(strsplit(Rversion, ".", fixed = TRUE)[[1L]][1L:2L], collapse = ".")
  } else if (is.list(Rversion)) {
    paste(Rversion$major, Rversion$minor, sep = ".")
  } else {
    Rversion
  }
}

# R.version
R.version_3.6.0 <- structure(
  list(
    platform = "x86_64-w64-mingw32",
    arch = "x86_64",
    os = "mingw32",
    system = "x86_64, mingw32",
    status = "",
    major = "3",
    minor = "6.0",
    year = "2019",
    month = "04",
    day = "26",
    `svn rev` = "76424",
    language = "R",
    version.string = "R version 3.6.0 (2019-04-26)",
    nickname = "Planting of a Tree"
  ),
  class = "simple.list"
)

# getRversion()
getRversion_3.6.0 <- structure(
  list(
    c(3L, 6L, 0L)
  ),
  class = c("R_system_version", "package_version", "numeric_version")
)

expect_3.6 <- function(object) {
  expect_equal(object, "3.6")
}

test_that("twoDigitRversion() behaves", {
  expect_3.6(
    twodigitRversion(list(major = 3, minor = 6))
  )
  expect_3.6(
    twodigitRversion("3.6")
  )
  expect_3.6(
    twodigitRversion("3.6.0")
  )
  expect_3.6(
    twodigitRversion("3.6.1")
  )
  expect_3.6(
    twodigitRversion(R.version_3.6.0)
  )
  expect_3.6(
    twodigitRversion(getRversion_3.6.0)
  )
})

#' Obtains DESCRIPTION metadata from CRAN for each package.
#'
#' This is a wrapper around `tools::CRAN_package_db` and may be
#' deprecated in future versions of the package.
#'
#' @inheritParams pkgDep
#' @inheritParams makeRepo
#'
#' @export
#'
#' @example /inst/examples/example_getCranDescription.R
getCranDescription <- function(
  pkg,
  repos = getOption("repos"),
  type = "source",
  pkgs = pkgDep(pkg, repos = repos, type = type)
) {
  if (getRversion() >= "3.4.1") {
    pdb <- tools::CRAN_package_db()
    pdb[match(pkgs, pdb$Package), ]
  } else {
    msg <- "This function is not available in R-3.4.0 or earlier."
    stop(msg)
  }
}

#' Check for available package updates in a miniCRAN repo.
#'
#' `oldPackages()` indicates packages which have a (suitable) later version on
#' the repositories whereas `updatePackages()` offers to download and install
#' such packages.
#'
#' These functions are based on [update.packages()].  However, rather than
#' looking for locally installed packages they look for the package source and
#' binaries in the miniCRAN repository.
#'
#' @name updatePackages
#'
#' @inheritParams makeRepo
#' @inheritParams pkgDep
#'
#' @param method  Download method, see [download.file()].
#'
#' @param availableLocal all packages hosted in the miniCRAN repo, as returned
#'   by [pkgAvail()]. A subset can be specified; currently this must be in the
#'   same (character matrix) format as returned by [pkgAvail()].
#'
#' @return `oldPackages()` returns a matrix with one row per package and columns
#'   for "Package", "LocalVer", "ReposVer" and "Repository".  The matrix row
#'   names the package names.
#'
#' @seealso [updatePackages()], [pkgAvail()].
#'
#' @export
#' @family update repo functions
#'
#' @example /inst/examples/example_updatePackages.R
#'
oldPackages <- function(
  path = NULL,
  repos = getOption("repos"),
  availPkgs = pkgAvail(repos = repos, type = type, Rversion = Rversion),
  method,
  availableLocal = pkgAvail(
    repos = path,
    type = type,
    Rversion = Rversion,
    quiet = quiet
  ),
  type = "source",
  Rversion = R.version,
  quiet = FALSE
) {
  if (is.null(path)) stop("path to miniCRAN repo must be specified")
  if (!missing(availPkgs)) {
    if (!is.matrix(availPkgs) || !is.character(availPkgs[, "Package"]))
      stop("ill-formed 'availPkgs' matrix")
  }
  if (NROW(availPkgs) == 0L) stop("Invalid remote repository")
  if (NROW(availableLocal) == 0L) stop("Invalid local repository")

  idx <- match(availableLocal[, "Package"], availPkgs[, "Package"])
  compare <- sapply(seq_along(idx), function(i) {
    compareVersion(
      (availPkgs[idx[i], "Version"]),
      (availableLocal[i, "Version"])
    ) >
      0
  })

  update <- cbind(
    availableLocal[compare, c("Package", "Version"), drop = FALSE],
    availPkgs[idx[compare], c("Version", "Repository"), drop = FALSE]
  )
  colnames(update) <- c("Package", "LocalVer", "ReposVer", "Repository")
  update
}


simplifyRepos <- function(repos, t, Rversion) {
  tail <- substring(contribUrl("---", type = t, Rversion = Rversion), 4)
  ind <- regexpr(tail, repos, fixed = TRUE)
  ind <- ifelse(ind > 0, ind - 1, nchar(repos, type = "c"))
  substr(repos, 1, ind)
}

graphics_capable <- function() {
  .Platform$OS.type == "windows" ||
    .Platform$GUI == "AQUA" ||
    (capabilities("tcltk") && capabilities("X11"))
}

select_from_list <- function(choices, preselect, multiple, title, graphics) {
  utils::select.list(choices, preselect, multiple, title, graphics)
}

ask_to_update <- function(oldPkgs, t, Rversion, ask = FALSE) {
  z <-
    if (is.character(ask) && ask == "graphics") {
      if (graphics_capable()) {
        k <- select_from_list(
          choices = oldPkgs[, 1L],
          preselect = oldPkgs[, 1L],
          multiple = TRUE,
          title = "Packages to be updated",
          graphics = TRUE
        )
        oldPkgs[match(k, oldPkgs[, 1L]), , drop = FALSE]
      } else {
        ask_to_update_package(oldPkgs, t, Rversion)
      }
    } else {
      if (isTRUE(ask)) {
        ask_to_update_package(oldPkgs, t, Rversion)
      } else {
        oldPkgs
      }
    }
  rownames(z) <- unname(z[, "Package"])
  z
}

read_line_wrapper <- function(prompt = "") {
  base::readline(prompt)
}

ask_to_update_package <- function(old, t, Rversion) {
  update <- NULL
  for (k in seq_len(nrow(old))) {
    cat(
      old[k, "Package"],
      ":\n",
      "Local Version",
      old[k, "LocalVer"],
      "\n",
      "Repos Version",
      old[k, "ReposVer"],
      "available at",
      simplifyRepos(old[k, "Repository"], t, Rversion)
    )
    cat("\n")
    answer <- tolower(substr(read_line_wrapper("Update (y/N/c)?  "), 1L, 1L))
    if (answer == "c") {
      cat("cancelled by user\n")
      return(invisible())
    }
    if (answer == "y") {
      update <- rbind(update, old[k, ])
    }
  }
  update
}


#' @inheritParams makeRepo
#'
#' @param oldPkgs if specified as non-NULL, `updatePackages()` only considers
#'   these packages for updating. This may be a character vector of package
#'   names or a matrix as returned by `oldPackages()`.
#'
#' @param ask logical indicating whether to ask user before packages are
#'   actually downloaded and installed.  Alternatively, the value `"graphics"`
#'   starts an interactive widget to allow the user to (de-)select from the list of
#'   packages which could be updated or added. The latter value only works on
#'   systems with a GUI version of [select.list()], and is otherwise equivalent
#'   to `ask = TRUE`.
#'
#' @return `updatePackages` returns `NULL` invisibly.
#'
#' @export
#'
updatePackages <- function(
  path = NULL,
  repos = getOption("repos"),
  method = NULL,
  ask = TRUE,
  availPkgs = pkgAvail(repos = repos, type = type, Rversion = Rversion),
  oldPkgs = NULL,
  type = "source",
  Rversion = R.version,
  quiet = FALSE
) {
  assert_that(is_path(path))

  do_one <- function(t, Rversion, ask) {
    if (!is.matrix(oldPkgs) && is.character(oldPkgs)) {
      subset <- oldPkgs
      oldPkgs <- NULL
    } else {
      subset <- NULL
    }
    if (is.null(oldPkgs)) {
      oldPkgs <- oldPackages(
        path = path,
        repos = repos,
        method = method,
        availPkgs = availPkgs,
        type = t,
        Rversion = Rversion
      )
      if (is.null(oldPkgs)) {
        message("All packages are up to date from repos: ", names(repos))
        return(invisible())
      }
    } else {
      if (!(is.matrix(oldPkgs) && is.character(oldPkgs))) {
        msg <- paste0(
          "invalid 'oldPkgs'; ",
          "must be a character vector or a result from oldPackages()"
        )
        stop(msg)
      }
    }
    if (!is.null(subset)) {
      oldPkgs <- oldPkgs[rownames(oldPkgs) %in% subset, , drop = FALSE]
      if (nrow(oldPkgs) == 0) return(invisible())
    }

    update <- ask_to_update(
      oldPkgs = oldPkgs,
      t = t,
      Rversion = Rversion,
      ask = ask
    )

    if (length(update[, "Package"])) {
      addPackage(
        update[, "Package"],
        path = path,
        repos = repos,
        type = t,
        quiet = quiet,
        deps = FALSE,
        Rversion = Rversion
      )
    }
  }

  lapply(type, do_one, Rversion = Rversion, ask = ask)
}

old <- matrix(
  c(
    "nomnoml",
    "BH",
    "digest",
    "2.1",
    "1.54.0-4",
    "0.6.4",
    "2.2",
    "1.55.0-1",
    "0.6.7",
    "https://packagemanager.posit.co/cran/2024-01-02",
    "https://packagemanager.posit.co/cran/2024-01-02",
    "https://packagemanager.posit.co/cran/2024-01-02"
  ),
  ncol = 4,
  dimnames = list(
    c("nomnoml", "BH", "digest"),
    c("Package", "LocalVer", "ReposVer", "Repository")
  )
)


test_that("asking for confirmation", {
  skip_if_not_installed("mockr")

  mockr::with_mock(
    read_line_wrapper = function(...) {
      "y"
    },
    .env = "miniCRAN",
    {
      capture.output(
        z <-
          ask_to_update(old, "source", "3.1", ask = TRUE)
      )
      expect_equal(old, z, ignore_attr = TRUE)
    }
  )

  mockr::with_mock(
    read_line_wrapper = function(...) {
      "n"
    },
    .env = "miniCRAN",
    {
      capture.output(
        z <-
          ask_to_update(old, "source", "3.1", ask = TRUE)
      )
      expect_null(z)
    }
  )

  mockr::with_mock(
    read_line_wrapper = function(...) {
      "y"
    },
    graphics_capable = function() {
      TRUE
    },
    select_from_list = function(choices, ...) {
      choices
    },
    .env = "miniCRAN",
    {
      capture.output(
        z <-
          ask_to_update(old, "source", "3.1", ask = "graphics")
      )
      expect_equal(old, z, ignore_attr = TRUE)
    }
  )
})

#' Returns names of base packages.
#'
#' Retrieves names of installed packages by calling
#' [utils::installed.packages()] and returning only those packages where
#' `Priority == "base"`.
#'
#' @export
#' @family dependency functions
#'
#' @seealso [pkgDep()]
basePkgs <- function()
  names(which(installed.packages()[, "Priority"] == "base"))

availPkgNames <- function(pdb) {
  pdb[, "Package"]
}

#' Retrieves package dependencies.
#'
#' Performs recursive retrieve for `Depends`, `Imports` and `LinkLibrary`.
#' Performs non-recursive retrieve for `Suggests`.
#'
#'
#' @param pkg Character vector of packages.
#'
#' @param availPkgs Data frame with an element called `package`. The `package`
#'   element is a vector of available packages.  Defaults to reading this list
#'   from CRAN, using [available.packages()]
#'
#' @param repos URL(s) of the 'contrib' sections of the repositories, e.g.
#'   `"https://cran.us.r-project.org"`. Passed to [available.packages()]
#'
#' @param type Possible values are (currently) "source", "mac.binary" and
#'   "win.binary": the binary types can be listed and downloaded but not
#'   installed on other platforms.  Passed to [download.packages()].
#'
#' @param depends If TRUE, retrieves `Depends`, `Imports` and `LinkingTo` dependencies
#'   (non-recursively)
#'
#' @param suggests If TRUE, retrieves Suggests dependencies (non-recursively)
#'
#' @param enhances If TRUE, retrieves Enhances dependencies (non-recursively)
#'
#' @param quiet If TRUE, suppresses warnings
#'
#' @param includeBasePkgs If TRUE, include base R packages in results
#'
#' @template Rversion
#'
#' @param ... Other arguments passed to [available.packages()]
#'
#' @export
#' @family dependency functions
#'
#' @return character vector of package names
#'
#' @example /inst/examples/example_pkgDep.R
#'
pkgDep <- function(
  pkg,
  availPkgs,
  repos = getOption("repos"),
  type = "source",
  depends = TRUE,
  suggests = TRUE,
  enhances = FALSE,
  includeBasePkgs = FALSE,
  Rversion = R.version,
  quiet = FALSE,
  ...
) {
  assert_that(is_package(pkg))

  if (!depends & !suggests & !enhances) {
    warning(
      "Returning nothing, since depends, suggests and enhances are all FALSE"
    )
    return(character(0))
  }

  if (missing(availPkgs)) {
    if (!is.null(names(repos)) & repos["CRAN"] == "@CRAN@") {
      repos <- p3m()
    }
    if (is.na(type)) type <- "source"
    availPkgs <- pkgAvail(
      repos = repos,
      type = type,
      Rversion = Rversion,
      quiet = quiet,
      ...
    )
  }

  assert_that(
    nrow(availPkgs) > 0,
    msg = sprintf("Unable to retrieve %s package %s from %s", type, pkg, repos)
  )

  pkgInAvail <- pkg %in% availPkgs[, "Package"]
  pkgInAvail <- pkg %in% availPkgNames(availPkgs)
  if (sum(pkgInAvail) == 0) stop("No valid packages in pkg")
  if (sum(pkgInAvail) < length(pkg)) {
    warning(
      "Package not recognized: ",
      paste(pkg[!pkgInAvail], collapse = ", ")
    )
  }

  n_req <- pkg[pkgInAvail]
  n_req_all <- pkg

  # Suggests
  if (suggests) {
    p_sug <- tools::package_dependencies(
      n_req,
      availPkgs,
      which = "Suggests",
      recursive = FALSE
    )
    n_sug <- unique(unname(unlist(p_sug)))
    n_req_all <- c(n_req_all, n_sug)
  } else {
    p_sug <- NA
  }

  # Enhances
  if (enhances) {
    p_enh <- tools::package_dependencies(
      n_req,
      availPkgs,
      which = "Enhances",
      recursive = FALSE
    )
    n_enh <- unique(unname(unlist(p_enh)))
    n_req_all <- c(n_req_all, n_enh)
  } else {
    p_enh <- NA
  }

  # Depends, Imports and LinkingTo
  p_dep <- tools::package_dependencies(
    n_req_all,
    availPkgs,
    which = c("Depends", "Imports", "LinkingTo"),
    recursive = TRUE
  )
  n_dep <- unique(unname(unlist(p_dep)))

  p_all <- p_dep
  n_all <- unique(c(n_dep, n_req_all))
  n_all <- c(n_req, setdiff(n_all, n_req))

  ret <- n_all
  if (!includeBasePkgs) ret <- ret[!ret %in% basePkgs()]
  attr(ret, "pkgs") <- list(
    n_req = n_req,
    n_all = n_all,
    p_dep = p_dep,
    p_sug = p_sug,
    p_enh = p_enh,
    p_all = p_all
  )
  class(ret) <- c("pkgDep", "character")
  assert_that(is_package_vector(ret))
  ret
}

#' @export
print.pkgDep <- function(x, ...) {
  attr(x, "pkgs") <- NULL
  class(x) <- "character"
  print(as.vector(x), ...)
}


#' Reads available packages from CRAN repository.
#'
#' This is a thin wrapper around [utils::available.packages()].  If the argument
#' `path` is supplied, then the function attempts to read from a local
#' repository, otherwise attempts to read from a CRAN mirror at the `repos` url.
#'
#' @inheritParams pkgDep
#'
#' @param filters passed to [utils::available.packages]
#'
#' @export
#' @family create repo functions
#' @seealso [pkgDep()]
pkgAvail <- function(
  repos = getOption("repos"),
  type = "source",
  Rversion = R.version,
  quiet = FALSE,
  filters = NULL
) {
  assert_that(is_repos(repos))
  if (!grepl("^https*://|file:///", repos[1]) && file.exists(repos[1])) {
    repos <- paste0(
      "file:///",
      normalizePath(repos[1], mustWork = FALSE, winslash = "/")
    )
  } else {
    if (!is.null(names(repos)) && isTRUE(unname(repos["CRAN"]) == "@CRAN@")) {
      repos <- p3m()
    }
  }
  ap <- function() {
    if (getRversion() >= "3.3.0") {
      utils::available.packages(
        contribUrl(repos, type = type, Rversion = Rversion),
        type = type,
        filters = filters,
        repos = repos
      )
    } else {
      utils::available.packages(
        contribUrl(repos, type = type, Rversion = Rversion),
        type = type,
        filters = filters
      )
    }
  }
  if (quiet) suppressWarnings(ap()) else ap()
}


# Modified copy of utils::contrib.url()
contribUrl <- function(
  repos = getOption("repos"),
  type = getOption("pkgType"),
  Rversion = R.version
) {
  Rversion <- twodigitRversion(Rversion)
  if (type == "both") type <- "source"
  if (type == "binary") type <- .Platform$pkgType
  if (is.null(repos)) return(NULL)

  if ("@CRAN@" %in% repos && interactive()) {
    cat(
      gettext("--- Please select a CRAN mirror for use in this session ---"),
      "\n",
      sep = ""
    )
    flush.console()
    chooseCRANmirror()
    m <- match("@CRAN@", repos)
    nm <- names(repos)
    repos[m] <- getOption("repos")["CRAN"]
    if (is.null(nm)) nm <- rep("", length(repos))
    nm[m] <- "CRAN"
    names(repos) <- nm
  }
  if ("@CRAN@" %in% repos) stop("trying to use CRAN without setting a mirror")
  ver <- Rversion
  mac.path <- "macosx"
  if (substr(type, 1L, 11L) == "mac.binary.") {
    mac.path <- paste(mac.path, substring(type, 12L), sep = "/")
    type <- "mac.binary"
  }
  res <- switch(
    type,
    source = paste(gsub("/$", "", repos), "src", "contrib", sep = "/"),
    mac.binary = paste(
      gsub("/$", "", repos),
      "bin",
      mac.path,
      "contrib",
      ver,
      sep = "/"
    ),
    win.binary = paste(
      gsub("/$", "", repos),
      "bin",
      "windows",
      "contrib",
      ver,
      sep = "/"
    )
  )
  res
}

if (interactive()) library(testthat)

# incorrect package list --------------------------------------------------

test_that("pkgDep throws warnings and errors", {
  expect_error(
    pkgDep(availPkgs = cranJuly2014),
    "argument \"pkg\" is missing, with no default"
  )

  expect_error(
    pkgDep(matrix(1:3, 3), availPkgs = cranJuly2014),
    "pkg should be a character vector with package names"
  )

  expect_error(
    pkgDep("reshape99", availPkgs = cranJuly2014),
    "No valid packages in pkg"
  )

  expect_error(
    pkgDep(c("reshape98", "reshape99"), availPkgs = cranJuly2014),
    "No valid packages in pkg"
  )

  expect_warning(
    pkgDep(c("reshape2", "reshape99"), availPkgs = cranJuly2014),
    "Package not recognized: reshape99"
  )
})


# suggests ----------------------------------------------------------------

test_that("pkgDep treats suggests correctly", {
  exp <- pkgDep("ggplot2", availPkgs = cranJuly2014, suggests = FALSE)
  expect_s3_class(exp, "pkgDep")
  expect_identical(
    as.vector(exp),
    c(
      "ggplot2",
      "plyr",
      "digest",
      "gtable",
      "reshape2",
      "scales",
      "proto",
      "MASS",
      "Rcpp",
      "stringr",
      "RColorBrewer",
      "dichromat",
      "munsell",
      "labeling",
      "colorspace"
    )
  )

  exp <- pkgDep("ggplot2", availPkgs = cranJuly2014, suggests = TRUE)
  expect_s3_class(exp, "pkgDep")
  expect_identical(
    as.vector(exp),
    c(
      "ggplot2",
      "plyr",
      "digest",
      "gtable",
      "reshape2",
      "scales",
      "proto",
      "MASS",
      "Rcpp",
      "stringr",
      "RColorBrewer",
      "dichromat",
      "munsell",
      "labeling",
      "colorspace",
      "SparseM",
      "lattice",
      "survival",
      "Formula",
      "latticeExtra",
      "cluster",
      "maps",
      "sp",
      "foreign",
      "mvtnorm",
      "TH.data",
      "sandwich",
      "zoo",
      "evaluate",
      "formatR",
      "highr",
      "markdown",
      "mime",
      "nlme",
      "Matrix",
      "quantreg",
      "Hmisc",
      "mapproj",
      "hexbin",
      "maptools",
      "multcomp",
      "testthat",
      "knitr",
      "mgcv"
    )
  )
})


# includeBasePkgs ---------------------------------------------------------

test_that("pkgDep treats includeBasePkgs correctly", {
  exp <- pkgDep(
    "reshape2",
    includeBasePkgs = TRUE,
    availPkgs = cranJuly2014,
    suggests = FALSE
  )
  expect_s3_class(exp, "pkgDep")
  expect_identical(
    as.vector(exp),
    c("reshape2", "plyr", "stringr", "Rcpp", "methods")
  )

  exp <- pkgDep(
    "reshape2",
    includeBasePkgs = FALSE,
    availPkgs = cranJuly2014,
    suggests = FALSE
  )
  expect_s3_class(exp, "pkgDep")
  expect_identical(
    as.vector(exp),
    c("reshape2", "plyr", "stringr", "Rcpp")
  )
})

if (getRversion() >= "3.1.0") {
  utils::globalVariables(c("."))
}

#' Check for previous versions of packages in a miniCRAN repository.
#'
#' Checks for previous versions, and returns the file paths for packages with
#' multiple versions. You can subsequently decide which version to keep.
#'
#' @param pkgs Character vector of packages to be installed. If not provided,
#'   checks all files for multiple package versions.
#'
#' @param path The local path to the directory where the miniCRAN repo resides.
#'
#' @param type  character, indicating the type of package to download and
#'   install. See [install.packages()].
#'
#' @template Rversion
#'
#' @return Returns invisibly the file paths to packages with multiple versions
#'   for removal.
#'
#' @export
#'
#' @return list with an element for each `type`, consisting of a character
#'   vector of download paths
#'
#' @family update repo functions
#'
#' @example /inst/examples/example_checkVersions.R
#'
checkVersions <- function(
  pkgs = NULL,
  path = NULL,
  type = "source",
  Rversion = R.version
) {
  if (is.null(path)) stop("path must be specified.")
  if (!file.exists(path)) stop("invalid path, ", path)

  do_one <- function(type) {
    pkgPath <- repoBinPath(path, type, Rversion)
    if (is.null(pkgs)) {
      files <- dir(pkgPath)
    } else {
      files <- sapply(
        pkgs,
        function(x) list.files(pkgPath, pattern = paste0("^", x, "_"))
      )
    }
    files <- unlist(files)
    pkgFiles <- grep("\\.(tar\\.gz|zip|tgz)$", basename(files), value = TRUE)

    # identify duplicate packages and warn the user
    pkgs <- sapply(strsplit(files, "_"), "[[", 1)
    dupes <- pkgs[duplicated(pkgs)]
    if (length(dupes)) {
      warning(
        "Duplicate package(s): ",
        paste(dupes, collapse = ", "),
        call. = FALSE
      )
    }
    file.path(pkgPath, pkgFiles)
  }

  duplicatePkgs <- sapply(type, do_one, simplify = FALSE)
  names(duplicatePkgs) <- type
  duplicatePkgs
}


# addPackage---------------------------------------------------------------

#' Add packages to a miniCRAN repository.
#'
#' @inheritParams makeRepo
#' @inheritParams pkgDep
#'
#' @template Rversion
#'
#' @param deps logical indicating whether the package dependencies should be
#'   added (default `TRUE`).
#'
#' @return Installs the packages, rebuilds the package index, and invisibly
#'   returns the number of packages written to the index files.
#'
#' @importFrom tools write_PACKAGES
#' @export
#' @family update repo functions
#'
#' @example /inst/examples/example_checkVersions.R
#'
addPackage <- function(
  pkgs = NULL,
  path = NULL,
  repos = getOption("repos"),
  type = "source",
  Rversion = R.version,
  writePACKAGES = TRUE,
  deps = TRUE,
  quiet = FALSE
) {
  if (is.null(path) || is.null(pkgs))
    stop("path and pkgs must both be specified.")

  do_one <- function(t) {
    prev <- checkVersions(
      pkgs = pkgs,
      path = path,
      type = t,
      Rversion = Rversion
    )
    prev <- prev[[1]]
    prev.df <- getPkgVersFromFile(prev)

    if (deps) pkgs <- pkgDep(pkgs, repos = repos, type = t, Rversion = Rversion)

    makeRepo(
      pkgs = pkgs,
      path = path,
      repos = repos,
      type = t,
      Rversion = Rversion,
      download = TRUE,
      writePACKAGES = FALSE,
      quiet = quiet
    )

    if (length(prev)) {
      curr <- suppressWarnings(
        checkVersions(pkgs = pkgs, path = path, type = t, Rversion = Rversion)
      )
      curr <- curr[[1]]
      curr.df <- getPkgVersFromFile(curr)

      findPrevPackage <- function(x) {
        grep(paste0("^", x), basename(prev))
      }

      dupes <- with(curr.df, package[duplicated(package)])
      if (length(dupes)) {
        to_remove <- lapply(dupes, findPrevPackage)
        if (length(unlist(to_remove))) {
          file.remove(prev[unlist(to_remove)])
        }
      }
    }
  }

  ret <- lapply(type, do_one)

  if (writePACKAGES)
    updateRepoIndex(path = path, type = type, Rversion = Rversion)
  invisible(ret)
}


# addOldPackage -----------------------------------------------------------

#' Add old package versions to a miniCRAN repository.
#'
#' Will download and add older source package versions. Older binary versions
#' are not normally available on CRAN and should be built from source on the
#' platform for which they are required. As such, specifying `type!="source"`
#' will likely fail as the download will not be successful.
#'
#' @inheritParams addPackage
#' @inheritParams makeRepo
#'
#' @param vers The package version to install.
#'
#' @return Adds the packages, rebuilds the package index, and invisibly returns
#'   the number of packages written to the index files.
#'
#' @note Dependencies for old package versions cannot be determined
#'   automatically and must be specified by the user in `pkgs` and `vers`. Thus,
#'   `deps=FALSE` is the default for this function.
#'
#' @importFrom tools write_PACKAGES
#' @export
#' @family update repo functions
#'
#' @example /inst/examples/example_checkVersions.R
#'
addOldPackage <- function(
  pkgs = NULL,
  path = NULL,
  vers = NULL,
  repos = getOption("repos"),
  type = "source",
  Rversion = R.version,
  writePACKAGES = TRUE,
  deps = FALSE,
  quiet = TRUE
) {
  if (is.null(path) || is.null(pkgs) || is.null(vers)) {
    stop("path, pkgs, and vers must all be specified.")
  }
  if (type != "source")
    stop(
      "Older binary versions are not normally available on CRAN. ",
      "You must build the binary versions from source."
    )
  if (deps) {
    message(
      "Unable to automatically determine dependency version information.\n",
      "Use pkgs and vers to identify which dependecies and their versions to download."
    )
  }
  vers <- as.character(vers)
  oldPkgs <- file.path(
    repos,
    repoPrefix(type, R.version),
    "Archive",
    pkgs,
    sprintf("%s_%s%s", pkgs, vers, pkgFileExt(type))
  )

  pkgPath <- repoBinPath(path = path, type = type, Rversion = Rversion)
  if (!file.exists(pkgPath)) dir.create(pkgPath, recursive = TRUE)

  do_one <- function(x) {
    result <- download.file(
      x,
      destfile = file.path(pkgPath, basename(x)),
      method = "auto",
      mode = "wb",
      quiet = quiet
    )
    if (result != 0) warning("error downloading file ", x)
  }
  ret <- sapply(oldPkgs, do_one)
  if (writePACKAGES) {
    updateRepoIndex(path = path, type = type, Rversion)
  }
  invisible(ret)
}


# .listFiles -------------------------------------------------------------------

#' List pre-built packages in a directory based on file extension
#'
#' @param pkgs  Character vector of package names
#' @param path  Character string specifying the directory containing packages to
#'   be added.
#' @param type  Character indicating the package type (e.g., "source",
#'   "win.binary", etc.).
#'
#' @return Installs the packages and returns the new package index.
#'
#' @rdname listFiles
#' @docType methods
#'
#' @keywords Internal
#'
#' @examples
#' \dontrun{
#'  .listFiles('path/to/my/packages', type = "source")
#' }
#'
.listFiles <- function(pkgs, path, type) {
  stopifnot(dir.exists(path))
  pattern <- pkgFileExt(type)

  # get a list of all files in pkgPaths directory matching pattern
  f <- list.files(path, pattern = pattern)

  assert_that(
    is.character(f) && length(f) > 0,
    msg = sprintf("No files found in path with extension '%s'", pattern)
  )

  # we only care about the subset matching pkgs
  f <- sapply(pkgs, function(x) {
    grep(x, f, value = TRUE)
  })

  if (length(f)) {
    # if multiple versions present, always use latest
    fp <- local({
      x <- strsplit(f, "_")
      sapply(x, `[[`, 1)
    })

    fv <- local({
      x <- strsplit(f, "_")
      x <- sapply(x, `[[`, 2)
      x <- strsplit(x, pattern)
      x <- sapply(x, `[[`, 1)
      as.numeric_version(x)
    })

    fout <- sapply(fp, function(x) {
      ids.p <- which(fp %in% x)

      # numeric_version always returns version using '.' as separator,
      #   even if the package uses '-', so we need to ensure either will work
      id.v <- which(fv == max(fv[ids.p]))

      f[id.v]
    })
    unique(fout)
  } else {
    character()
  }
}


#' Add local packages to a miniCRAN repository.
#'
#' Examine the contents of a directory specified by `pkgPath` for pre-built
#' packages matching the names specified by `pkgs`, and add these to the
#' miniCRAN repository.
#'
#' To build a package from source and then add it, use `build = TRUE`. Note that
#' package development libraries and the `devtools` package must be installed on
#' your system in order to build packages.
#'
#' @note Currently, adding local packages does not check nor download their
#'   dependencies.
#'
#' @inheritParams addPackage
#'
#' @param deps Not used. See note.
#'
#' @param pkgPath  Character vector of directory location containing packages to
#'   be added. Note that `pkgPath` should be the parent directory of the package
#'   (i.e., the package directory path is constructed from `file.path(pkgPath,
#'   pkgs)`).
#'
#' @param build Logical indicating whether packages should be build prior to
#'   adding.
#'
#' @return Installs the packages and returns the new package index.
#'
#' @export
#' @docType methods
#'
#' @author Alex Chubaty
#'
#' @examples
#' \dontrun{
#'  addLocalPackage("myPackage", "path/to/my/prebuilt/package",
#'                  "path/to/my/miniCRAN/repo")
#'
#'  addLocalPackage("myPackage", "path/to/my/package/sourcecode",
#'                  "path/to/my/miniCRAN/repo", build = TRUE)
#' }
#'
addLocalPackage <- function(
  pkgs = NULL,
  pkgPath = NULL,
  path = NULL,
  type = "source",
  Rversion = R.version,
  writePACKAGES = TRUE,
  deps = FALSE,
  quiet = FALSE,
  build = FALSE
) {
  if (is.null(path) || is.null(pkgs) || is.null(pkgPath)) {
    stop("path, pkgs, and pkgPath must be specified.")
  }

  stopifnot(dir.exists(file.path(pkgPath)))

  # build local package if needed
  if (isTRUE(build)) {
    warning("Building local packages is still being tested.")
    if (requireNamespace("devtools", quietly = TRUE)) {
      lapply(pkgs, function(x) {
        devtools::build(
          pkg = file.path(pkgPath, x),
          path = pkgPath,
          binary = ifelse(type == "source", FALSE, TRUE),
          quiet = quiet
        )
      })
    } else {
      stop("To build packages, you must first install the 'devtools' package.")
    }
  }

  # get list of pre-built packages for each type, filter by pkgs to be added
  do_one <- function(t) {
    repoPath <- file.path(path, repoPrefix(t, Rversion))
    if (!dir.exists(repoPath)) dir.create(repoPath, recursive = TRUE)
    files <- .listFiles(pkgs = pkgs, path = pkgPath, type = t)

    # check for previous package version and omit if identical
    prev <- checkVersions(pkgs, path)
    same <- which(basename(as.character(prev)) %in% files)

    if (length(same)) {
      files <- files[-same]
      if (length(files) == 0) {
        if (!quiet) message("All packages up to date. Nothing to add.")
        return(invisible(NULL))
      }
    }

    # copy other packages to their respective folders
    lapply(files, function(x) {
      f.src <- file.path(pkgPath, x)
      f.dst <- file.path(repoPath, x)

      file.exists(f.src)

      if (!isTRUE(quiet)) message("copying ", x, "\n")
      file.copy(from = f.src, to = f.dst)
    })

    # check to ensure they all copied successfully
    copied <- file.exists(file.path(repoPath, files))
    if (!all(copied)) {
      warning(
        "the following ",
        t,
        " packages were not copied:\n",
        paste(files[!copied], sep = ", ")
      )
    }

    # remove previous package versions
    if (length(prev[-same]) > 0) unlink(prev[-same])
    file.path(repoPath, files)[copied]
  }

  ret <- sapply(type, do_one)

  # write package index for each folder:
  if (writePACKAGES) {
    updateRepoIndex(path = path, type = type, Rversion = Rversion)
  }
  invisible(ret)
}

#' description
#'
#' At the end of 2014, CRAN consisted of more than 6,000 packages.  Many
#' organisations need to maintain a private mirror of CRAN, but with only a
#' subset of packages that are relevant to them.
#'
#' `miniCRAN` makes it possible to create an internally consistent repository
#' consisting of selected packages from CRAN-like repositories.  The user
#' specifies a set of desired packages, and miniCRAN recursively reads the
#' dependency tree for these packages, then downloads only this subset.
#'
#'
#' There are many reasons for not creating a complete mirror CRAN using `rsync`:
#'
#' * You may wish to mirror only a subset of CRAN, for security, legal
#' compliance or any other in-house reason
#'
#' * You may wish to restrict internal package use to a subset of public
#' packages, to minimize package duplication, or other reasons of coding
#' standards
#'
#' * You may wish to make packages available from public repositories other than
#' CRAN, e.g. BioConductor, r-forge, OmegaHat, etc.
#'
#' * You may wish to add custom in-house packages to your repository
#'
#' The ambition of `miniCRAN` is to eventually satisfy all of these
#' considerations.
#'
#' @section Making a private repo:
#'
#'   * [pkgAvail()]: Read from a local (or remote) CRAN-like repository and
#'   determine available packages.
#'
#'   * [pkgDep()]: Find (recursive) package dependencies.
#'
#'   * [makeRepo()] : Make a mini CRAN repository, by downloading packages (and
#'   their dependencies) and creating the appropriate file structure for a
#'   repository.  This allows you to use functions like
#'   [utils::available.packages()] and [utils::install.packages()] on your local
#'   repository.
#'
#'   This subset will be internally consistent, i.e. the following functions
#'   will work as expected:
#'
#'   * [utils::available.packages()]
#'
#'   * [utils::install.packages()]
#'
#'   The main function is [makeRepo()] - this will download all the required
#'   packages, with their dependencies, into the appropriate repository file
#'   structure, and then create the repository index (PACKAGES) file.
#'
#'
#'
#' @section Updating packages in a repo:
#'
#'   * [oldPackages()]: Indicates packages which have a (suitable) later version
#'   on the repositories * [updatePackages()]: Offers to download and install
#'   such packages
#'
#'
#'
#' @section Creating dependencies:
#'
#'   To get a recursive list of dependencies as well as a plot, use [pkgDep()]
#'   followed by  [makeDepGraph()].
#'
#'   * [pkgDep()]: Find (recursive) package dependencies.
#'
#'   * [makeDepGraph()]: Create graph of selected package dependencies.
#'
#'   * [plot.pkgDepGraph()]: Create a visualization of the dependency graph
#'
#'
#'
#' @section Package options:
#'
#'   \describe{ \item{\code{minicran.mran}}{preferred p3m URL. Defaults to
#'   \url{https://packagemanager.posit.co/cran} for R versions 3.2.2 and greater. Versions
#'   earlier than 3.2.2 use HTTP instead of HTTPS.} }
#'
#' @name miniCRAN-package
#' @aliases miniCRAN minicran
#' @docType package
#' @keywords package
#' @importFrom graphics legend par title
#' @importFrom stats reshape setNames
#' @importFrom utils chooseCRANmirror compareVersion download.file flush.console
#'   installed.packages select.list
"_PACKAGE"


# data documentation ------------------------------------------------------

#' Stored version of available.packages()
#'
#' Copy of the result of [utils::available.packages()] on July 1, 2014.
#'
#' @docType data
#' @keywords datasets
#' @name cranJuly2014
#' @usage cranJuly2014
#' @format matrix
NULL

---
title:  "Using repositories other than CRAN with miniCRAN"
author: "Andrie de Vries"
date:   "`r as.character(format(Sys.Date(), format = '%B %d, %Y'))`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Using repositories other than CRAN with miniCRAN}
  %\VignettePackage{miniCRAN}
  %\VignetteKeyword{miniCRAN}
  %\usepackage[utf8]{inputenc}
---

Although the package name `miniCRAN` seems to indicate you can only use CRAN as a repository, you can in fact use any CRAN-like repository.

This vignette contains some examples of how to refer to different package repositories, including CRAN, alternative mirrors of CRAN, R-Forge as well as BioConductor.

To simplify the code to show the salient features, we use a little helper function, `index()` that is a simple wrapper around `available.packages()`:


```{r setup}
# Wrapper around available.packages ---------------------------------------
 
index <- function(url, type = "source", filters = NULL, head = 5, 
                  cols = c("Package", "Version")) {
  contribUrl <- contrib.url(url, type = type)
  p <- available.packages(contribUrl, type = type, filters = filters)
  p[1:head, cols]
}
 
```

## Using CRAN

The URL for the master mirror in Austria:

```{r CRAN, eval=FALSE}
CRAN <- "https://cran.r-project.org"
index(CRAN)
```

Your output should look like this (albeit with different packages and version numbers):

```
              Package         Version
A3            "A3"            "1.0.0"
AalenJohansen "AalenJohansen" "1.0"  
AATtools      "AATtools"      "0.0.2"
ABACUS        "ABACUS"        "1.0.0"
abasequence   "abasequence"   "0.1.0"
```

## Using a different mirror

You can also point to any other mirror, for example the snapshot taken on 2024-01-02 by Posit Public Package Manager:

```{r p3m, eval=FALSE}
p3m <- "https://packagemanager.posit.co/cran/2024-01-02"
index(p3m)
```

## Using R-forge

R-forge has CRAN-like structure:

```{r rforge, eval=FALSE}
rforge <- "https://r-forge.r-project.org"
index(rforge)
```

---
title:  "Using miniCRAN to identify package dependencies"
author: "Andrie de Vries"
date:   "`r as.character(format(Sys.Date(), format='%B %d, %Y'))`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Using miniCRAN to identify package dependencies}
  %\VignettePackage{miniCRAN}
  %\VignetteKeyword{miniCRAN}
  %\usepackage[utf8]{inputenc}
--- 

The `miniCRAN` package exposes two functions that provide information about dependencies:

- The function `pkgDep()` returns a character vector with the names of dependencies. Internally, `pkgDep()` is a wrapper around `tools::package_dependencies()`, a base R function that, well, tells you about package dependencies. My `pkgDep()` function is in one way a convenience, but more importantly it sets different defaults (more about this later).

- The function `makeDepGraph()` creates a graph representation of the dependencies.

The package `chron` neatly illustrates the different roles of Imports, Suggests and Enhances:

- `chron` **Imports** the base packages graphics and stats. This means that `chron` internally makes use of graphics and stats and will always load these packages.

- `chron` **Suggests** the packages scales and ggplot2. This means that `chron` uses some functions from these packages in examples or in its vignettes.  However, these functions are not necessary to use `chron`

- `chron` **Enhances** the package `zoo`, meaning that it adds something to `zoo` packages. These enhancements are made available to you if you have `zoo` installed.


## A worked example using the package chron

The function `pkgDep()` exposes not only these dependencies, but also all recursive dependencies. In other words, it answers the question which packages need to be installed to satisfy all dependencies of dependencies.

This means that the algorithm is as follows:

* First retrieve a list of `Suggests` and `Enhances`, using a non-recursive dependency search
* Next, perform a recursive search for all `Imports`, `Depends` and `LinkingTo`

The resulting list of packages should then contain the complete list necessary to satisfy all dependencies. In code:

```{r init}
library("miniCRAN")
```



```{r pkgdep}
tags <- "chron"
pkgDep(tags, availPkgs = cranJuly2014)
```


To create an igraph plot of the dependencies, use the function `makeDepGraph()` and plot the results: 

```{r makeDepGraph, warning=FALSE}
dg <- makeDepGraph(tags, enhances = TRUE, availPkgs = cranJuly2014)
set.seed(1)
plot(dg, legendPosition = c(-1, 1), vertex.size = 20)
```


Note how the dependencies expand to `zoo` (enhanced), `scales` and `ggplot` (suggested) and then recursively from there to get all the `Imports` and `LinkingTo` dependencies.



## An example with multiple input packages

As a final example, create a dependency graph of seven very popular R packages:

```{r so-tags, warning=FALSE, fig.width=10, fig.height=10}
tags <- c("ggplot2", "data.table", "plyr", "knitr", "shiny", "xts", "lattice")
pkgDep(tags, suggests = TRUE, enhances = FALSE, availPkgs = cranJuly2014)

dg <- makeDepGraph(tags, enhances = TRUE, availPkgs = cranJuly2014)
set.seed(1)
plot(dg, legendPosition = c(-1, -1), vertex.size = 10, cex = 0.7)
```

---
title: "Using miniCRAN to create and maintain a local CRAN repository"
author: "Andrie de Vries and Alex Chubaty"
date:   "`r as.character(format(Sys.Date(), format='%B %d, %Y'))`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using miniCRAN to create and maintain a local CRAN repository}
  %\VignettePackage{miniCRAN}
  %\VignetteKeyword{miniCRAN}
  %\usepackage[utf8]{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

Start by creating the recursive dependency tree for your target packages.

For example, imagine a scenario where you want to create a repository that consists of the package `foreach` and its dependencies.
 
Start by creating the dependency list:

```{r make-repo-1}
library("miniCRAN")

# define CRAN mirror
mirror <- c(CRAN = "https://cloud.r-project.org")

# Specify list of packages to download
pkgs <- c("foreach")
pkgList <- pkgDep(pkgs, repos = mirror, type = "source", suggests = FALSE, 
                  availPkgs = cranJuly2014)
pkgList
```

Next, create a repository with the function `makeRepo()`.  In this example, get the required files for `source` packages as well as windows binaries:

```{r make-repo-2, eval=FALSE}
# Create temporary folder for miniCRAN
dir.create(pth <- file.path(tempdir(), "miniCRAN"))

# Make repo for source and win.binary
makeRepo(pkgList, path = pth, repos = mirror, type = c("source", "win.binary"))
```


Investigate the repository file structure:

```{r make-repo-3, eval=FALSE}
# List all files in miniCRAN
list.files(pth, recursive = TRUE, full.names = FALSE)
```


Use `pkgAvail` to list available packages in your repository:

```{r make-repo-4, eval=FALSE}
# Check for available packages
pkgAvail(repos = pth, type = "win.binary")[, c(1:3, 5)]
```

# Install packages from your local repository

To install packages from a local repository, you need to use the [Uniform Resource Identifier](https://en.wikipedia.org/wiki/Uniform_Resource_Identifier) (URI) convention `file:///` to point to your file location.

```{r make-repo-5, eval=FALSE}
install.packages(pkgs, 
                 repos = paste0("file:///", pth),
                 type = "source")
```


# Adding packages to an existing miniCRAN repository

## Adding new packages from CRAN

After creating a local miniCRAN repository, additional packages and their dependencies can easily be added. This mechanism can also be used to re-add an existing package to the miniCRAN repo.

```{r addto-repo-new-1, eval=FALSE}
# Add new packages (from CRAN) to the miniCRAN repo
addPackage("Matrix", path = pth, repos = mirror, type = c("source", "win.binary"))
pkgAvail(repos = pth, type = "win.binary")[, c(1:3, 5)]
```

The value that is returned (invisibly) via `addPackage` is the number of packages written to the index file, *i.e.*, the total number of packages in the repo of that type.

## Adding an older version of a package from CRAN

To add a specific version of a package from CRAN (or another CRAN-like repository), we can easily download the source packages from the CRAN archives. Dependencies for old package versions cannot be determined automatically and must be specified by the user.

*Note:* in order to to add binaries of older packages, you will need to download the source and build the binaries on the intended platform yourself. You will need the appropriate R development tools installed in order to build package binaries from source.

```{r addto-repo-old-1, eval=FALSE}
# create a data frame with the package and version info
oldVers <- data.frame(
  package = c("foreach", "codetools", "iterators"),
  version = c("1.4.0", "0.2-7", "1.0.5"),
  stringsAsFactors = FALSE
)

# download old source package version and create repo index
addOldPackage(pkgList, path = pth, vers = oldVers$version, repos = mirror, type = "source")
```

You will get a warning whenever there are multiple versions of a package saved in the repository. Currently, you need to manually remove duplicate versions before rebuilding the repository's package index.

Note: This last step is important, otherwise you may end up with a repo in an inconsistent state.

```{r addto-repo-old-2, eval=FALSE}
# List package versions in the miniCRAN repo (produces warning about duplicates)
pkgVersionsSrc <- checkVersions(pkgList, path = pth, type = "source")
pkgVersionsBin <- checkVersions(pkgList, path = pth, type = "win.binary")

# After inspecting package versions, remove old versions
basename(pkgVersionsSrc) # duplicate versions
basename(pkgVersionsBin)

file.remove(pkgVersionsSrc[c(2,4,6)])

# rebuild the package index after removing duplicate package versions
updateRepoIndex(pth, type = c("source", "win.binary"))
```

To see the updated list of packages available in the miniCRAN repo:

```{r addto-repo-old-3, eval=FALSE}
pkgAvail(pth, type = "source")[, c(1:3, 5)] # contains the old versions
pkgAvail(pth, type = "win.binary")[, c(1:3, 5)] # contains the current versions
```



## Adding packages from other sources

This feature will be implemented in a future release.


# Updating the packages in a miniCRAN repository

Checking for updated versions of the packages currently stored in the miniCRAN repository:

```{r update-repo-1, eval=FALSE}
# Check if updated packages are available
oldPackages(path = pth, repos = mirror, type = "source")[, 1:3] # should need update
oldPackages(path = pth, repos = mirror, type = "win.binary")[, 1:3] # should be current
```


Update the versions of the packages currently stored in the miniCRAN repository. By default, a prompt is given to confirm the update for each package. This prompt can be suppressed using `ask = FALSE`, which will update all packages. Be careful using this option if you want to keep certain packages at an older version.

```{r update-repo-2, eval=FALSE}
# Update available packages
updatePackages(path = pth, repos = mirror, type = "source", ask = FALSE) # should need update
updatePackages(path = pth, repos = mirror, type = "win.binary", ask = FALSE) # should be current
```




```{r cleanup, include=FALSE, eval=FALSE}
# Delete temporary folder
unlink(pth, recursive = TRUE)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{twodigitRversion}
\alias{twodigitRversion}
\title{Get a two-digit version of the R version}
\usage{
twodigitRversion(Rversion = R.version)
}
\arguments{
\item{Rversion}{Version of R (only used if \code{type} is not \code{source}.) Defaults to \link{R.version}, but this can be specified as any of the following formats:
\itemize{
\item a character string with the two digit R version, e.g. "3.1"
\item a list with components \code{major} and \code{minor}
\item the result of \code{\link[=getRversion]{getRversion()}}
\item the result of \link{R.version}
}}
}
\value{
A character string representing the two-digit R version.
}
\description{
Get a two-digit version of the R version
}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/P3M.R
\name{is.online}
\alias{is.online}
\title{Returns TRUE if the p3m URL can be accessed.}
\usage{
is.online(url = NULL, tryHttp = TRUE)
}
\arguments{
\item{url}{p3m url}

\item{tryHttp}{If TRUE, also attempts http URL, for compatibility with older versions of R}
}
\description{
Returns TRUE if the p3m URL can be accessed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/minicran-package.R
\docType{package}
\name{miniCRAN-package}
\alias{miniCRAN-package}
\alias{miniCRAN}
\alias{minicran}
\title{description}
\description{
At the end of 2014, CRAN consisted of more than 6,000 packages.  Many
organisations need to maintain a private mirror of CRAN, but with only a
subset of packages that are relevant to them.
}
\details{
\code{miniCRAN} makes it possible to create an internally consistent repository
consisting of selected packages from CRAN-like repositories.  The user
specifies a set of desired packages, and miniCRAN recursively reads the
dependency tree for these packages, then downloads only this subset.

There are many reasons for not creating a complete mirror CRAN using \code{rsync}:
\itemize{
\item You may wish to mirror only a subset of CRAN, for security, legal
compliance or any other in-house reason
\item You may wish to restrict internal package use to a subset of public
packages, to minimize package duplication, or other reasons of coding
standards
\item You may wish to make packages available from public repositories other than
CRAN, e.g. BioConductor, r-forge, OmegaHat, etc.
\item You may wish to add custom in-house packages to your repository
}

The ambition of \code{miniCRAN} is to eventually satisfy all of these
considerations.
}
\section{Making a private repo}{

\itemize{
\item \code{\link[=pkgAvail]{pkgAvail()}}: Read from a local (or remote) CRAN-like repository and
determine available packages.
\item \code{\link[=pkgDep]{pkgDep()}}: Find (recursive) package dependencies.
\item \code{\link[=makeRepo]{makeRepo()}} : Make a mini CRAN repository, by downloading packages (and
their dependencies) and creating the appropriate file structure for a
repository.  This allows you to use functions like
\code{\link[utils:available.packages]{utils::available.packages()}} and \code{\link[utils:install.packages]{utils::install.packages()}} on your local
repository.
}

This subset will be internally consistent, i.e. the following functions
will work as expected:
\itemize{
\item \code{\link[utils:available.packages]{utils::available.packages()}}
\item \code{\link[utils:install.packages]{utils::install.packages()}}
}

The main function is \code{\link[=makeRepo]{makeRepo()}} - this will download all the required
packages, with their dependencies, into the appropriate repository file
structure, and then create the repository index (PACKAGES) file.
}

\section{Updating packages in a repo}{

\itemize{
\item \code{\link[=oldPackages]{oldPackages()}}: Indicates packages which have a (suitable) later version
on the repositories * \code{\link[=updatePackages]{updatePackages()}}: Offers to download and install
such packages
}
}

\section{Creating dependencies}{


To get a recursive list of dependencies as well as a plot, use \code{\link[=pkgDep]{pkgDep()}}
followed by  \code{\link[=makeDepGraph]{makeDepGraph()}}.
\itemize{
\item \code{\link[=pkgDep]{pkgDep()}}: Find (recursive) package dependencies.
\item \code{\link[=makeDepGraph]{makeDepGraph()}}: Create graph of selected package dependencies.
\item \code{\link[=plot.pkgDepGraph]{plot.pkgDepGraph()}}: Create a visualization of the dependency graph
}
}

\section{Package options}{


\describe{ \item{\code{minicran.mran}}{preferred p3m URL. Defaults to
\url{https://packagemanager.posit.co/cran} for R versions 3.2.2 and greater. Versions
earlier than 3.2.2 use HTTP instead of HTTPS.} }
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/andrie/miniCRAN}
  \item Report bugs at \url{https://github.com/andrie/miniCRAN/issues}
}

}
\author{
\strong{Maintainer}: Andrie de Vries \email{apdevries@gmail.com} [copyright holder]

Other contributors:
\itemize{
  \item Alex Chubaty \email{alex.chubaty@gmail.com} [contributor]
  \item Microsoft Corporation [copyright holder]
}

}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgDep.R
\name{basePkgs}
\alias{basePkgs}
\title{Returns names of base packages.}
\usage{
basePkgs()
}
\description{
Retrieves names of installed packages by calling
\code{\link[utils:installed.packages]{utils::installed.packages()}} and returning only those packages where
\code{Priority == "base"}.
}
\seealso{
\code{\link[=pkgDep]{pkgDep()}}

Other dependency functions: 
\code{\link{makeDepGraph}()},
\code{\link{pkgDep}()},
\code{\link{plot.pkgDepGraph}()}
}
\concept{dependency functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/minicran-package.R
\docType{data}
\name{cranJuly2014}
\alias{cranJuly2014}
\title{Stored version of available.packages()}
\format{
matrix
}
\usage{
cranJuly2014
}
\description{
Copy of the result of \code{\link[utils:available.packages]{utils::available.packages()}} on July 1, 2014.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/updatePackages.R
\name{updatePackages}
\alias{updatePackages}
\alias{oldPackages}
\title{Check for available package updates in a miniCRAN repo.}
\usage{
oldPackages(
  path = NULL,
  repos = getOption("repos"),
  availPkgs = pkgAvail(repos = repos, type = type, Rversion = Rversion),
  method,
  availableLocal = pkgAvail(repos = path, type = type, Rversion = Rversion, quiet =
    quiet),
  type = "source",
  Rversion = R.version,
  quiet = FALSE
)

updatePackages(
  path = NULL,
  repos = getOption("repos"),
  method = NULL,
  ask = TRUE,
  availPkgs = pkgAvail(repos = repos, type = type, Rversion = Rversion),
  oldPkgs = NULL,
  type = "source",
  Rversion = R.version,
  quiet = FALSE
)
}
\arguments{
\item{path}{Destination download path. This path is the root folder of your
new repository.}

\item{repos}{URL(s) of the 'contrib' sections of the repositories, e.g.
\code{"https://cran.us.r-project.org"}. Passed to \code{\link[=available.packages]{available.packages()}}}

\item{availPkgs}{Data frame with an element called \code{package}. The \code{package}
element is a vector of available packages.  Defaults to reading this list
from CRAN, using \code{\link[=available.packages]{available.packages()}}}

\item{method}{Download method, see \code{\link[=download.file]{download.file()}}.}

\item{availableLocal}{all packages hosted in the miniCRAN repo, as returned
by \code{\link[=pkgAvail]{pkgAvail()}}. A subset can be specified; currently this must be in the
same (character matrix) format as returned by \code{\link[=pkgAvail]{pkgAvail()}}.}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}

\item{Rversion}{Version of R (only used if \code{type} is not \code{source}.) Defaults to \link{R.version}, but this can be specified as any of the following formats:
\itemize{
\item a character string with the two digit R version, e.g. "3.1"
\item a list with components \code{major} and \code{minor}
\item the result of \code{\link[=getRversion]{getRversion()}}
\item the result of \link{R.version}
}}

\item{quiet}{If TRUE, suppress status messages (if any), and the progress bar
during download.}

\item{ask}{logical indicating whether to ask user before packages are
actually downloaded and installed.  Alternatively, the value \code{"graphics"}
starts an interactive widget to allow the user to (de-)select from the list of
packages which could be updated or added. The latter value only works on
systems with a GUI version of \code{\link[=select.list]{select.list()}}, and is otherwise equivalent
to \code{ask = TRUE}.}

\item{oldPkgs}{if specified as non-NULL, \code{updatePackages()} only considers
these packages for updating. This may be a character vector of package
names or a matrix as returned by \code{oldPackages()}.}
}
\value{
\code{oldPackages()} returns a matrix with one row per package and columns
for "Package", "LocalVer", "ReposVer" and "Repository".  The matrix row
names the package names.

\code{updatePackages} returns \code{NULL} invisibly.
}
\description{
\code{oldPackages()} indicates packages which have a (suitable) later version on
the repositories whereas \code{updatePackages()} offers to download and install
such packages.
}
\details{
These functions are based on \code{\link[=update.packages]{update.packages()}}.  However, rather than
looking for locally installed packages they look for the package source and
binaries in the miniCRAN repository.
}
\examples{
### `oldPackages` and `updatePackages` require an existing miniCRAN repo

# Specify list of packages to download
mirror <- c(CRAN = "https://cloud.r-project.org")
pkgs <- c("foreach")

pdb <- cranJuly2014

if (interactive()) {
  pdb <- pkgAvail(repos = mirror, type = "source")
  
  pkgList <- pkgDep(pkgs, availPkgs = pdb, repos = mirror, type = "source", suggests = FALSE)
  pkgList
  
  # Create temporary folder for miniCRAN
  dir.create(pth <- file.path(tempdir(), "miniCRAN"))
  
  # create the miniCRAN directory structure but only add bin files
  makeRepo(pkgList, path = pth, repos = mirror, type = "source", download = FALSE)
  makeRepo(pkgList, path = pth, repos = mirror, type = "win.binary", download = TRUE)
  
  # download old source package version and create repo index
  oldVers <- data.frame(package = c("foreach", "codetools", "iterators"),
                        version = c("1.4.0", "0.2-7", "1.0.5"),
                        stringsAsFactors = FALSE)
  addOldPackage(pkgList, path = pth, repos = mirror, vers = oldVers$version, type = "source")
  # NOTE: older binary versions would need to be build from source
  
  # Check if updated packages are available
  oldPackages(path = pth, repos = mirror, type = "source") # should need update
  oldPackages(path = pth, repos = mirror, type = "win.binary") # should be current
  
  # Update available packages
  updatePackages(path = pth, repos = mirror, type = "source", ask = FALSE) # should need update
  updatePackages(path = pth, repos = mirror, type = "win.binary") # should be current
  
  # Delete temporary folder
  unlink(pth, recursive = TRUE)
}
}
\seealso{
\code{\link[=updatePackages]{updatePackages()}}, \code{\link[=pkgAvail]{pkgAvail()}}.

Other update repo functions: 
\code{\link{addOldPackage}()},
\code{\link{addPackage}()},
\code{\link{checkVersions}()},
\code{\link{makeRepo}()}
}
\concept{update repo functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeRepo.R
\name{makeRepo}
\alias{makeRepo}
\alias{updateRepoIndex}
\title{Downloads packages from CRAN to specified path and creates a local
repository.}
\usage{
makeRepo(
  pkgs,
  path,
  repos = getOption("repos"),
  type = "source",
  Rversion = R.version,
  download = TRUE,
  writePACKAGES = TRUE,
  filters = NULL,
  quiet = FALSE
)

updateRepoIndex(path, type = "source", Rversion = R.version)
}
\arguments{
\item{pkgs}{Character vector of packages to download}

\item{path}{Destination download path. This path is the root folder of your
new repository.}

\item{repos}{URL(s) of the 'contrib' sections of the repositories, e.g.
\code{"https://cran.us.r-project.org"}. Passed to \code{\link[=available.packages]{available.packages()}}}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}

\item{Rversion}{Version of R (only used if \code{type} is not \code{source}.) Defaults to \link{R.version}, but this can be specified as any of the following formats:
\itemize{
\item a character string with the two digit R version, e.g. "3.1"
\item a list with components \code{major} and \code{minor}
\item the result of \code{\link[=getRversion]{getRversion()}}
\item the result of \link{R.version}
}}

\item{download}{If TRUE downloads packages.}

\item{writePACKAGES}{If TRUE, calls \code{\link[=write_PACKAGES]{write_PACKAGES()}} to update the
repository PACKAGES file.}

\item{filters}{passed to \link[utils:available.packages]{utils::available.packages}}

\item{quiet}{If TRUE, suppress status messages (if any), and the progress bar
during download.}
}
\value{
character vector of downloaded package files
}
\description{
Given a list of packages, downloads these packages to a specified destination
folder using the required CRAN folder structure, and finally creates the
PACKAGES index file.  Since the folder structure mimics the required
structure and files of a CRAN repository, it supports functions like
\code{\link[utils:install.packages]{utils::install.packages()}}.
}
\note{
Internally makes use of \code{\link[utils:download.packages]{utils::download.packages()}} and
\code{\link[=write_PACKAGES]{write_PACKAGES()}}
}
\section{Repo folder structure}{


A repository has two main folders, one for source packages, and the other
for binary packages.  Inside the binary package folder, \code{bin}, you will find
subfolders for Windows as well as the various OSX binaries.

\code{+- Root}

\code{...+- src/contrib}

\code{......+- PACKAGES}

\code{..+- bin}

\code{.......+- windows/contrib/version}

\code{..........+- PACKAGES}

\code{.......+-  macosx/contrib/version}

\code{..........+-  PACKAGES}

\code{.......+-  macosx/mavericks/contrib/version}

\code{..........+-  PACKAGES}

\code{.......+-  macosx/leopard/contrib/version}

\code{..........+- PACKAGES}
}

\examples{

# Specify list of packages to download
mirror <- c(CRAN = "https://cloud.r-project.org")
pkgs <- c("foreach")

if (interactive()) {
  pdb <- cranJuly2014
  
  pdb <- pkgAvail(
    repos = c(CRAN = getOption("minicran.mran")),
    type = "source"
  )
  
  pkgList <- pkgDep(pkgs, availPkgs = pdb, repos = mirror,
                    type = "source", suggests = FALSE)
  pkgList
  
  
  # Create temporary folder for miniCRAN
  dir.create(pth <- file.path(tempdir(), "miniCRAN"))
  
  # Make repo for source and win.binary
  makeRepo(pkgList, path = pth, repos = mirror, type = "source")
  
  # List all files in miniCRAN
  list.files(pth, recursive = TRUE)
  
  # Check for available packages
  pkgAvail(repos = pth, type = "source")
  
  # Repeat process for windows binaries
  makeRepo(pkgList, path = pth, repos = mirror, type = "win.binary")
  pkgAvail(repos = pth, type = "win.binary")
  
  # Delete temporary folder
  unlink(pth, recursive = TRUE)
}
}
\seealso{
Other update repo functions: 
\code{\link{addOldPackage}()},
\code{\link{addPackage}()},
\code{\link{checkVersions}()},
\code{\link{updatePackages}()}
}
\concept{update repo functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addPackages.R
\docType{methods}
\name{addLocalPackage}
\alias{addLocalPackage}
\title{Add local packages to a miniCRAN repository.}
\usage{
addLocalPackage(
  pkgs = NULL,
  pkgPath = NULL,
  path = NULL,
  type = "source",
  Rversion = R.version,
  writePACKAGES = TRUE,
  deps = FALSE,
  quiet = FALSE,
  build = FALSE
)
}
\arguments{
\item{pkgs}{Character vector of packages to download}

\item{pkgPath}{Character vector of directory location containing packages to
be added. Note that \code{pkgPath} should be the parent directory of the package
(i.e., the package directory path is constructed from \code{file.path(pkgPath, pkgs)}).}

\item{path}{Destination download path. This path is the root folder of your
new repository.}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}

\item{Rversion}{Version of R (only used if \code{type} is not \code{source}.) Defaults to \link{R.version}, but this can be specified as any of the following formats:
\itemize{
\item a character string with the two digit R version, e.g. "3.1"
\item a list with components \code{major} and \code{minor}
\item the result of \code{\link[=getRversion]{getRversion()}}
\item the result of \link{R.version}
}}

\item{writePACKAGES}{If TRUE, calls \code{\link[=write_PACKAGES]{write_PACKAGES()}} to update the
repository PACKAGES file.}

\item{deps}{Not used. See note.}

\item{quiet}{If TRUE, suppress status messages (if any), and the progress bar
during download.}

\item{build}{Logical indicating whether packages should be build prior to
adding.}
}
\value{
Installs the packages and returns the new package index.
}
\description{
Examine the contents of a directory specified by \code{pkgPath} for pre-built
packages matching the names specified by \code{pkgs}, and add these to the
miniCRAN repository.
}
\details{
To build a package from source and then add it, use \code{build = TRUE}. Note that
package development libraries and the \code{devtools} package must be installed on
your system in order to build packages.
}
\note{
Currently, adding local packages does not check nor download their
dependencies.
}
\examples{
\dontrun{
 addLocalPackage("myPackage", "path/to/my/prebuilt/package",
                 "path/to/my/miniCRAN/repo")

 addLocalPackage("myPackage", "path/to/my/package/sourcecode",
                 "path/to/my/miniCRAN/repo", build = TRUE)
}

}
\author{
Alex Chubaty
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{repoPrefix}
\alias{repoPrefix}
\title{Get the path to the repo directory containing the package files.}
\usage{
repoPrefix(type, Rversion)
}
\arguments{
\item{type}{character, indicating the type of package to download and
install. See \code{\link[=install.packages]{install.packages()}}.}

\item{Rversion}{Version of R (only used if \code{type} is not \code{source}.) Defaults to \link{R.version}, but this can be specified as any of the following formats:
\itemize{
\item a character string with the two digit R version, e.g. "3.1"
\item a list with components \code{major} and \code{minor}
\item the result of \code{\link[=getRversion]{getRversion()}}
\item the result of \link{R.version}
}}
}
\value{
The file path to the package files directory.
}
\description{
Get the path to the repo directory containing the package files.
}
\note{
Not all versions of R are compatible with with all package types (e.g.,
\code{mac.binary.el-capitan} is only valid for R > 3.4.0).
}
\section{Repo folder structure}{


A repository has two main folders, one for source packages, and the other
for binary packages.  Inside the binary package folder, \code{bin}, you will find
subfolders for Windows as well as the various OSX binaries.

\code{+- Root}

\code{...+- src/contrib}

\code{......+- PACKAGES}

\code{..+- bin}

\code{.......+- windows/contrib/version}

\code{..........+- PACKAGES}

\code{.......+-  macosx/contrib/version}

\code{..........+-  PACKAGES}

\code{.......+-  macosx/mavericks/contrib/version}

\code{..........+-  PACKAGES}

\code{.......+-  macosx/leopard/contrib/version}

\code{..........+- PACKAGES}
}

\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgDep.R
\name{pkgAvail}
\alias{pkgAvail}
\title{Reads available packages from CRAN repository.}
\usage{
pkgAvail(
  repos = getOption("repos"),
  type = "source",
  Rversion = R.version,
  quiet = FALSE,
  filters = NULL
)
}
\arguments{
\item{repos}{URL(s) of the 'contrib' sections of the repositories, e.g.
\code{"https://cran.us.r-project.org"}. Passed to \code{\link[=available.packages]{available.packages()}}}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}

\item{Rversion}{Version of R (only used if \code{type} is not \code{source}.) Defaults to \link{R.version}, but this can be specified as any of the following formats:
\itemize{
\item a character string with the two digit R version, e.g. "3.1"
\item a list with components \code{major} and \code{minor}
\item the result of \code{\link[=getRversion]{getRversion()}}
\item the result of \link{R.version}
}}

\item{quiet}{If TRUE, suppresses warnings}

\item{filters}{passed to \link[utils:available.packages]{utils::available.packages}}
}
\description{
This is a thin wrapper around \code{\link[utils:available.packages]{utils::available.packages()}}.  If the argument
\code{path} is supplied, then the function attempts to read from a local
repository, otherwise attempts to read from a CRAN mirror at the \code{repos} url.
}
\seealso{
\code{\link[=pkgDep]{pkgDep()}}
}
\concept{create repo functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addPackages.R
\name{addOldPackage}
\alias{addOldPackage}
\title{Add old package versions to a miniCRAN repository.}
\usage{
addOldPackage(
  pkgs = NULL,
  path = NULL,
  vers = NULL,
  repos = getOption("repos"),
  type = "source",
  Rversion = R.version,
  writePACKAGES = TRUE,
  deps = FALSE,
  quiet = TRUE
)
}
\arguments{
\item{pkgs}{Character vector of packages to download}

\item{path}{Destination download path. This path is the root folder of your
new repository.}

\item{vers}{The package version to install.}

\item{repos}{URL(s) of the 'contrib' sections of the repositories, e.g.
\code{"https://cran.us.r-project.org"}. Passed to \code{\link[=available.packages]{available.packages()}}}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}

\item{Rversion}{Version of R (only used if \code{type} is not \code{source}.) Defaults to \link{R.version}, but this can be specified as any of the following formats:
\itemize{
\item a character string with the two digit R version, e.g. "3.1"
\item a list with components \code{major} and \code{minor}
\item the result of \code{\link[=getRversion]{getRversion()}}
\item the result of \link{R.version}
}}

\item{writePACKAGES}{If TRUE, calls \code{\link[=write_PACKAGES]{write_PACKAGES()}} to update the
repository PACKAGES file.}

\item{deps}{logical indicating whether the package dependencies should be
added (default \code{TRUE}).}

\item{quiet}{If TRUE, suppress status messages (if any), and the progress bar
during download.}
}
\value{
Adds the packages, rebuilds the package index, and invisibly returns
the number of packages written to the index files.
}
\description{
Will download and add older source package versions. Older binary versions
are not normally available on CRAN and should be built from source on the
platform for which they are required. As such, specifying \code{type!="source"}
will likely fail as the download will not be successful.
}
\note{
Dependencies for old package versions cannot be determined
automatically and must be specified by the user in \code{pkgs} and \code{vers}. Thus,
\code{deps=FALSE} is the default for this function.
}
\examples{
### `checkVersions` and `add.packages.miniCRAN` require an existing miniCRAN repo

# Specify list of packages to download
mirror <- c(CRAN = "https://cloud.r-project.org")
mirror
pkgs <- c("foreach")
pkgTypes <- c("source", "win.binary")

if (interactive()) {
  if (!is.online()) {
    message("p3m seems to be not available.  Check your internet connection.")
  } else {
    pdb <- pkgAvail(repos = mirror, type = "source")
  }
} else {
  pdb <- cranJuly2014
}


if (interactive()) {
  if (!is.online()) {
    message("p3m seems to be not available.  Check your internet connection.")
  } else {
    pkgList <- pkgDep(pkgs, availPkgs = pdb, repos = mirror, type = "source", suggests = FALSE)
    pkgList
  }
}

# Create temporary folder for miniCRAN

if (interactive()) {
  if (!is.online()) {
    message("p3m seems to be not available.  Check your internet connection.")
  } else {
    dir.create(pth <- file.path(tempdir(), "miniCRAN"))
    
    # Make repo for source and win.binary
    makeRepo(pkgList, path = pth, repos = mirror, type = pkgTypes)
    
    # Add other versions of a package (and assume these were added previously)
    oldVers <- data.frame(
      package = c("foreach", "codetools", "iterators"),
      version = c("1.4.0", "0.2-7", "1.0.5"),
      stringsAsFactors = FALSE
    )
    pkgs <- oldVers$package
    addOldPackage(pkgs, path = pth, vers = oldVers$version, repos = mirror, type = "source")
    # NOTE: older binary versions would need to be build from source
    
    # List package versions in the miniCRAN repo (produces warning about duplicates)
    pkgVersionsSrc <- checkVersions(pkgs, path = pth, type = "source")
    pkgVersionsBin <- checkVersions(pkgs, path = pth, type = "win.binary")
    
    # After inspecting package versions, remove old versions
    basename(pkgVersionsSrc$source) # "foreach_1.4.0.tar.gz"  "foreach_1.4.2.tar.gz"
    basename(pkgVersionsBin$win.binary) # "foreach_1.4.0.zip"     "foreach_1.4.2.zip"
    file.remove(c(pkgVersionsSrc$source[1], pkgVersionsBin$win.binary[1]))
    
    # Rebuild package index after adding/removing files
    updateRepoIndex(pth, type = pkgTypes, Rversion = R.version)
    
    pkgAvail(pth, type = "source")
    
    # Add new packages (from CRAN) to the miniCRAN repo
    addPackage("Matrix", path = pth, repos = mirror, type = pkgTypes)
    
    # Delete temporary folder
    unlink(pth, recursive = TRUE)
  }
}
}
\seealso{
Other update repo functions: 
\code{\link{addPackage}()},
\code{\link{checkVersions}()},
\code{\link{makeRepo}()},
\code{\link{updatePackages}()}
}
\concept{update repo functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addPackages.R
\docType{methods}
\name{.listFiles}
\alias{.listFiles}
\title{List pre-built packages in a directory based on file extension}
\usage{
.listFiles(pkgs, path, type)
}
\arguments{
\item{pkgs}{Character vector of package names}

\item{path}{Character string specifying the directory containing packages to
be added.}

\item{type}{Character indicating the package type (e.g., "source",
"win.binary", etc.).}
}
\value{
Installs the packages and returns the new package index.
}
\description{
List pre-built packages in a directory based on file extension
}
\examples{
\dontrun{
 .listFiles('path/to/my/packages', type = "source")
}

}
\keyword{Internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addPackages.R
\name{addPackage}
\alias{addPackage}
\title{Add packages to a miniCRAN repository.}
\usage{
addPackage(
  pkgs = NULL,
  path = NULL,
  repos = getOption("repos"),
  type = "source",
  Rversion = R.version,
  writePACKAGES = TRUE,
  deps = TRUE,
  quiet = FALSE
)
}
\arguments{
\item{pkgs}{Character vector of packages to download}

\item{path}{Destination download path. This path is the root folder of your
new repository.}

\item{repos}{URL(s) of the 'contrib' sections of the repositories, e.g.
\code{"https://cran.us.r-project.org"}. Passed to \code{\link[=available.packages]{available.packages()}}}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}

\item{Rversion}{Version of R (only used if \code{type} is not \code{source}.) Defaults to \link{R.version}, but this can be specified as any of the following formats:
\itemize{
\item a character string with the two digit R version, e.g. "3.1"
\item a list with components \code{major} and \code{minor}
\item the result of \code{\link[=getRversion]{getRversion()}}
\item the result of \link{R.version}
}}

\item{writePACKAGES}{If TRUE, calls \code{\link[=write_PACKAGES]{write_PACKAGES()}} to update the
repository PACKAGES file.}

\item{deps}{logical indicating whether the package dependencies should be
added (default \code{TRUE}).}

\item{quiet}{If TRUE, suppress status messages (if any), and the progress bar
during download.}
}
\value{
Installs the packages, rebuilds the package index, and invisibly
returns the number of packages written to the index files.
}
\description{
Add packages to a miniCRAN repository.
}
\examples{
### `checkVersions` and `add.packages.miniCRAN` require an existing miniCRAN repo

# Specify list of packages to download
mirror <- c(CRAN = "https://cloud.r-project.org")
mirror
pkgs <- c("foreach")
pkgTypes <- c("source", "win.binary")

if (interactive()) {
  if (!is.online()) {
    message("p3m seems to be not available.  Check your internet connection.")
  } else {
    pdb <- pkgAvail(repos = mirror, type = "source")
  }
} else {
  pdb <- cranJuly2014
}


if (interactive()) {
  if (!is.online()) {
    message("p3m seems to be not available.  Check your internet connection.")
  } else {
    pkgList <- pkgDep(pkgs, availPkgs = pdb, repos = mirror, type = "source", suggests = FALSE)
    pkgList
  }
}

# Create temporary folder for miniCRAN

if (interactive()) {
  if (!is.online()) {
    message("p3m seems to be not available.  Check your internet connection.")
  } else {
    dir.create(pth <- file.path(tempdir(), "miniCRAN"))
    
    # Make repo for source and win.binary
    makeRepo(pkgList, path = pth, repos = mirror, type = pkgTypes)
    
    # Add other versions of a package (and assume these were added previously)
    oldVers <- data.frame(
      package = c("foreach", "codetools", "iterators"),
      version = c("1.4.0", "0.2-7", "1.0.5"),
      stringsAsFactors = FALSE
    )
    pkgs <- oldVers$package
    addOldPackage(pkgs, path = pth, vers = oldVers$version, repos = mirror, type = "source")
    # NOTE: older binary versions would need to be build from source
    
    # List package versions in the miniCRAN repo (produces warning about duplicates)
    pkgVersionsSrc <- checkVersions(pkgs, path = pth, type = "source")
    pkgVersionsBin <- checkVersions(pkgs, path = pth, type = "win.binary")
    
    # After inspecting package versions, remove old versions
    basename(pkgVersionsSrc$source) # "foreach_1.4.0.tar.gz"  "foreach_1.4.2.tar.gz"
    basename(pkgVersionsBin$win.binary) # "foreach_1.4.0.zip"     "foreach_1.4.2.zip"
    file.remove(c(pkgVersionsSrc$source[1], pkgVersionsBin$win.binary[1]))
    
    # Rebuild package index after adding/removing files
    updateRepoIndex(pth, type = pkgTypes, Rversion = R.version)
    
    pkgAvail(pth, type = "source")
    
    # Add new packages (from CRAN) to the miniCRAN repo
    addPackage("Matrix", path = pth, repos = mirror, type = pkgTypes)
    
    # Delete temporary folder
    unlink(pth, recursive = TRUE)
  }
}
}
\seealso{
Other update repo functions: 
\code{\link{addOldPackage}()},
\code{\link{checkVersions}()},
\code{\link{makeRepo}()},
\code{\link{updatePackages}()}
}
\concept{update repo functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.pkgDepGraph.R
\name{plot.pkgDepGraph}
\alias{plot.pkgDepGraph}
\title{Plots a package dependency graph.}
\usage{
\method{plot}{pkgDepGraph}(
  x,
  pkgsToHighlight,
  main = paste(attr(x, "pkgs"), collapse = ", "),
  legendPosition = c(-1.2, -1),
  shape = "circle",
  vertex.size = 8,
  cex = 1,
  ...
)
}
\arguments{
\item{x}{Object to plot}

\item{pkgsToHighlight}{Optional character vector with names of package to
highlight. If missing, defaults to packages used in original call to
\code{\link[=makeDepGraph]{makeDepGraph()}}}

\item{main}{Title of plot}

\item{legendPosition}{Numeric vector of length 2, indicating (x, y) position
of edge legend. Both values should be in the range \verb{[-1; 1]}.  If \code{NULL},
the edge legend is not displayed.}

\item{shape}{Shape of edge.  See \code{\link[igraph:plot.common]{igraph::igraph.plotting()}}. Could be
"none", "circle", "square", ...}

\item{vertex.size}{Size of vertex shape. \code{\link[igraph:plot.common]{igraph::igraph.plotting()}}}

\item{cex}{Vertex label size.}

\item{...}{Ignored}
}
\description{
Plots a package dependency graph.
}
\examples{
tags <- "chron"

# Plot using defaults

if (interactive()){
  pdb <- pkgAvail(
    repos = c(CRAN = getOption("minicran.mran")),
    type = "source"
  )
} else {
  pdb <- cranJuly2014
}


if (interactive()) {
  dg <- makeDepGraph(tags, availPkgs = pdb  , includeBasePkgs = FALSE,
                     suggests = TRUE, enhances = TRUE)
  
  set.seed(43);
  plot(dg)
  
  
  # Move edge legend to top left
  set.seed(42);
  plot(dg, legendPosition = c(-1, 1))
  
  # Change font size and shape size
  set.seed(42);
  plot(dg, legendPosition = c(-1, 1), vertex.size = 20,  cex = 0.5)
  
  
  # Move vertex legend to top right
  set.seed(42);
  plot(dg, legendPosition = c(1, 1), vertex.size = 20,  cex = 0.5)
}
}
\seealso{
Other dependency functions: 
\code{\link{basePkgs}()},
\code{\link{makeDepGraph}()},
\code{\link{pkgDep}()}
}
\concept{dependency functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github-functions.R
\name{addPackageListingGithub}
\alias{addPackageListingGithub}
\title{Add DESCRIPTION information from package on github.}
\usage{
addPackageListingGithub(
  pdb = pkgAvail(),
  repo,
  username = NULL,
  branch = "main"
)
}
\arguments{
\item{pdb}{Package database, usually the result of \code{\link[=pkgAvail]{pkgAvail()}} or
\code{\link[=available.packages]{available.packages()}}}

\item{repo}{Character vector. Name of repository on github, e.g.
\code{"andrie/rrd"}}

\item{username}{Optional character vector. Name of repository on github, e.g.
\code{"andrie/rrd"}}

\item{branch}{name of branch, defaults to \code{"main"}}
}
\description{
Downloads the DESCRIPTION file from a package on github, parses the fields
and adds (or replaces) a row in the available package database.
}
\examples{
# Create package database
pdb <- cranJuly2014

if (interactive()) {
  pdb <- pkgAvail(repos = c(CRAN = "https://cloud.r-project.org"))

  # Overwrite pdb with development version of miniCRAN at github
  newpdb <- addPackageListingGithub(pdb = pdb, "andrie/miniCRAN")
  newpdb["miniCRAN", ]

  # Add package from github that's not currently on CRAN
  newpdb <- addPackageListingGithub(pdb = pdb, repo = "tidyverse/ggplot2", branch = "main")
  newpdb["ggplot2", ]

  set.seed(1)
  plot(makeDepGraph("ggplot2", availPkgs = newpdb, suggests = TRUE))
}
}
\concept{github functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCranDescription.R
\name{getCranDescription}
\alias{getCranDescription}
\title{Obtains DESCRIPTION metadata from CRAN for each package.}
\usage{
getCranDescription(
  pkg,
  repos = getOption("repos"),
  type = "source",
  pkgs = pkgDep(pkg, repos = repos, type = type)
)
}
\arguments{
\item{pkg}{Character vector of packages.}

\item{repos}{URL(s) of the 'contrib' sections of the repositories, e.g.
\code{"https://cran.us.r-project.org"}. Passed to \code{\link[=available.packages]{available.packages()}}}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}

\item{pkgs}{Character vector of packages to download}
}
\description{
This is a wrapper around \code{tools::CRAN_package_db} and may be
deprecated in future versions of the package.
}
\examples{
if (interactive()) {
  getCranDescription(c("igraph", "ggplot2", "XML"),
                     repos = c(CRAN = getOption("minicran.mran"))
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgDep.R
\name{pkgDep}
\alias{pkgDep}
\title{Retrieves package dependencies.}
\usage{
pkgDep(
  pkg,
  availPkgs,
  repos = getOption("repos"),
  type = "source",
  depends = TRUE,
  suggests = TRUE,
  enhances = FALSE,
  includeBasePkgs = FALSE,
  Rversion = R.version,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{pkg}{Character vector of packages.}

\item{availPkgs}{Data frame with an element called \code{package}. The \code{package}
element is a vector of available packages.  Defaults to reading this list
from CRAN, using \code{\link[=available.packages]{available.packages()}}}

\item{repos}{URL(s) of the 'contrib' sections of the repositories, e.g.
\code{"https://cran.us.r-project.org"}. Passed to \code{\link[=available.packages]{available.packages()}}}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}

\item{depends}{If TRUE, retrieves \code{Depends}, \code{Imports} and \code{LinkingTo} dependencies
(non-recursively)}

\item{suggests}{If TRUE, retrieves Suggests dependencies (non-recursively)}

\item{enhances}{If TRUE, retrieves Enhances dependencies (non-recursively)}

\item{includeBasePkgs}{If TRUE, include base R packages in results}

\item{Rversion}{Version of R (only used if \code{type} is not \code{source}.) Defaults to \link{R.version}, but this can be specified as any of the following formats:
\itemize{
\item a character string with the two digit R version, e.g. "3.1"
\item a list with components \code{major} and \code{minor}
\item the result of \code{\link[=getRversion]{getRversion()}}
\item the result of \link{R.version}
}}

\item{quiet}{If TRUE, suppresses warnings}

\item{...}{Other arguments passed to \code{\link[=available.packages]{available.packages()}}}
}
\value{
character vector of package names
}
\description{
Performs recursive retrieve for \code{Depends}, \code{Imports} and \code{LinkLibrary}.
Performs non-recursive retrieve for \code{Suggests}.
}
\examples{

if (interactive()) {
  pkgDep(pkg = c("ggplot2", "plyr", "reshape2"),
         repos = c(CRAN = "https://cloud.r-project.org")
  )
  
  pdb <- cranJuly2014
  pdb <- pkgAvail(repos = c(CRAN = getOption("minicran.mran")))
  
  pkgDep(pkg = c("ggplot2", "plyr", "reshape2"), pdb)
  
}
}
\seealso{
Other dependency functions: 
\code{\link{basePkgs}()},
\code{\link{makeDepGraph}()},
\code{\link{plot.pkgDepGraph}()}
}
\concept{dependency functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeRepo.R
\name{makeLibrary}
\alias{makeLibrary}
\title{Deprecated function to download packages to local folder.}
\usage{
makeLibrary(pkgs, path, type = "source")
}
\arguments{
\item{pkgs}{Character vector of packages to download}

\item{path}{Destination download path. This path is the root folder of your
new repository.}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}
}
\description{
Deprecated function to download packages to local folder.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeDepGraph.R
\name{makeDepGraph}
\alias{makeDepGraph}
\title{Create dependency graph from available packages.}
\usage{
makeDepGraph(
  pkg,
  availPkgs,
  repos = getOption("repos"),
  type = "source",
  suggests = TRUE,
  enhances = FALSE,
  includeBasePkgs = FALSE,
  ...
)
}
\arguments{
\item{pkg}{Character vector of packages.}

\item{availPkgs}{Data frame with an element called \code{package}. The \code{package}
element is a vector of available packages.  Defaults to reading this list
from CRAN, using \code{\link[=available.packages]{available.packages()}}}

\item{repos}{URL(s) of the 'contrib' sections of the repositories, e.g.
\code{"https://cran.us.r-project.org"}. Passed to \code{\link[=available.packages]{available.packages()}}}

\item{type}{Possible values are (currently) "source", "mac.binary" and
"win.binary": the binary types can be listed and downloaded but not
installed on other platforms.  Passed to \code{\link[=download.packages]{download.packages()}}.}

\item{suggests}{If TRUE, retrieves Suggests dependencies (non-recursively)}

\item{enhances}{If TRUE, retrieves Enhances dependencies (non-recursively)}

\item{includeBasePkgs}{If TRUE, include base R packages in results}

\item{...}{Other arguments passed to \code{\link[=available.packages]{available.packages()}}}
}
\description{
Each package is a node, and a dependency is an edge
}
\examples{

if (interactive()) {
  availPkgs <- cranJuly2014
  
  availPkgs <- pkgAvail(
    repos = c(CRAN = "https://cloud.r-project.org"),
    type = "source"
  )
  
  
  # Create dependency graph using stored database of available packages
  p <- makeDepGraph(
    c("ggplot2", "forecast"),
    availPkgs = availPkgs
  )
  
  if(require(igraph)) plot(p)

  # Create dependency graph using newly retrieved database from CRAN
  
  p <- makeDepGraph(
    c("ggplot2", "forecast"),
    repos = c(CRAN = getOption("minicran.mran")),
    type = "source"
  )
  if(requireNamespace("igraph", quietly = TRUE)) {
    plot(p)
  } else {
    message("install package `igraph` to view dependency graph")
  }
  
}
}
\seealso{
\code{\link[=pkgDep]{pkgDep()}} to extract package dependencies

Other dependency functions: 
\code{\link{basePkgs}()},
\code{\link{pkgDep}()},
\code{\link{plot.pkgDepGraph}()}
}
\concept{dependency functions}
